# -*- encoding:utf-8 -*-
"""
中间层，从上层拿到x，y，df
拥有create estimator

"""

import os, functools
from enum import Enum

import numpy as np
import pandas as pd

from sklearn.base import TransformerMixin, ClassifierMixin, RegressorMixin, clone
from sklearn import metrics
from sklearn.datasets import load_iris
from sklearn.feature_selection import RFE, VarianceThreshold
from sklearn.preprocessing import label_binarize, StandardScaler, binarize

from ultron.ump.core.fixes import train_test_split, cross_val_score, mean_squared_error_scorer, six, signature
from ultron.optimize.grirdsearch import execute as ml_execute
from ultron.optimize.grirdsearch.creater import MLCreater
from ultron.kdutils.decorator import params_to_numpy
from ultron.kdutils.progress import Progress
from ultron.utilities.logger import kd_logger


class _EMLScoreType(Enum):
    """针对有监督学习的度量支持enum"""
    """有监督学习度量准确率"""
    E_SCORE_ACCURACY = 'accuracy'
    """有监督学习度量mse"""
    E_SCORE_MSE = mean_squared_error_scorer
    """有监督学习度量roc_auc"""
    E_SCORE_ROC_AUC = 'roc_auc'


class EMLFitType(Enum):
    """支持常使用的学习器类别enum"""
    """有监督学习：自动选择，根据y的label数量，> 10使用回归否则使用分类"""
    E_FIT_AUTO = 'auto'
    """有监督学习：回归"""
    E_FIT_REG = 'reg'
    """有监督学习：分类"""
    E_FIT_CLF = 'clf'
    """无监督学习：HMM"""
    E_FIT_HMM = 'hmm'
    """无监督学习：PCA"""
    E_FIT_PCA = 'pca'
    """无监督学习：KMEAN"""
    E_FIT_KMEAN = 'kmean'


def entry_wrapper(support=(EMLFitType.E_FIT_CLF, EMLFitType.E_FIT_REG,
                           EMLFitType.E_FIT_HMM, EMLFitType.E_FIT_PCA,
                           EMLFitType.E_FIT_KMEAN)):
    """
    类装饰器函数，对关键字参数中的fiter_type进行标准化，eg，fiter_type参数是'clf'， 转换为EMLFitType(fiter_type)
    赋予self.fiter_type，检测当前使用的具体学习器不在support参数中不执行被装饰的func函数了，打个log返回

    :param support: 默认 support=(EMLFitType.E_FIT_CLF, EMLFitType.E_FIT_REG, EMLFitType.E_FIT_HMM,
                           EMLFitType.E_FIT_PCA, EMLFitType.E_FIT_KMEAN)
                    即支持所有，被装饰的函数根据自身特性选择装饰参数
    """

    def decorate(func):

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            org_fiter_type = self.fiter_type
            if 'fiter_type' in kwargs:
                # 如果传递了fiter_type参数，pop出来
                fiter_type = kwargs.pop('fiter_type')
                # 如果传递的fiter_type参数是str，eg：'clf'， 转换为EMLFitType(fiter_type)
                if isinstance(fiter_type, six.string_types):
                    fiter_type = EMLFitType(fiter_type)
                self.fiter_type = fiter_type

            check_support = self.fiter_type
            if self.fiter_type == EMLFitType.E_FIT_AUTO:
                # 把auto的归到具体的分类或者回归
                check_y = self.y
                if 'y' in kwargs:
                    check_y = kwargs['y']
                check_support = EMLFitType.E_FIT_CLF if len(
                    np.unique(check_y)) <= 10 else EMLFitType.E_FIT_REG
            if check_support not in support:
                # 当前使用的具体学习器不在support参数中不执行被装饰的func函数了，打个log返回
                kd_logger.info('{} not support {}!'.format(
                    func.__name__, check_support.value))
                # 如果没能成功执行把类型再切换回来
                self.fiter_type = org_fiter_type
                return

            return func(self, *args, **kwargs)

        return wrapper

    return decorate


class AutoModel():
    """封装有简单学习及无监督学习方法以及相关操作类"""

    def __init__(self, x, y, df, fiter_type=EMLFitType.E_FIT_AUTO):
        """
        AbuML属于中间层需要所有原料都配齐，x, y, df，构造方式参考
        create_test_fiter方法中的实行流程

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param df: 拆分x，y使用的pd.DataFrame对象
        :param fiter_type: 使用的学习器类型，默认使用E_FIT_AUTO即根据y的label数量选择分类或者回归
        """
        self.estimator = MLCreater()
        # 如果传递进来的是字符串类型，转换为EMLFitType
        if isinstance(fiter_type, six.string_types):
            fiter_type = EMLFitType(fiter_type)
        self.x = x
        self.y = y
        self.df = df
        # ipython notebook下使用logging.info
        self.fiter_type = fiter_type

    def is_supervised_learning(self):
        """
        返回self.fiter_type所使用的是有监督学习还是无监督学习
        :return: bool，True: 有监督，False: 无监督
        """
        return self.fiter_type == EMLFitType.E_FIT_REG or self.fiter_type == EMLFitType.E_FIT_CLF or \
            self.fiter_type == EMLFitType.E_FIT_AUTO

    def echo_info(self, fiter=None):
        """
        显示fiter class信息，self.df信息包括，head，tail，describe
        eg：
            fiter class is: DecisionTreeClassifier(class_weight=None, criterion='gini', max_depth=None,
                max_features=None, max_leaf_nodes=None,
                min_impurity_split=1e-07, min_samples_leaf=1,
                min_samples_split=2, min_weight_fraction_leaf=0.0,
                presort=False, random_state=None, splitter='best')
            describe:
                          y        x0        x1        x2        x3
            count  150.0000  150.0000  150.0000  150.0000  150.0000
            mean     1.0000    5.8433    3.0540    3.7587    1.1987
            std      0.8192    0.8281    0.4336    1.7644    0.7632
            min      0.0000    4.3000    2.0000    1.0000    0.1000
            25%      0.0000    5.1000    2.8000    1.6000    0.3000
            50%      1.0000    5.8000    3.0000    4.3500    1.3000
            75%      2.0000    6.4000    3.3000    5.1000    1.8000
            max      2.0000    7.9000    4.4000    6.9000    2.5000
        :param fiter:
        :return:
        """
        if fiter is None:
            fiter = self.get_fiter()
        ('fiter class is: {}'.format(fiter))
        kd_logger.info('describe:\n{}'.format(self.df.describe()))
        kd_logger.info('head:\n{}'.format(self.df.head()))
        kd_logger.info('tail:\n{}'.format(self.df.tail()))

    def get_fiter(self):
        """
        根据self.fiter_type的类型选择从self.estimator返回学习器对象

        self.fiter_type == EMLFitType.E_FIT_AUTO：
            自动选择有简单学习，当y的label数量 < 10个使用分类self.estimator.clf，否则回归self.estimator.reg
        self.fiter_type == EMLFitType.E_FIT_REG:
            使用有监督学习回归self.estimator.reg
        self.fiter_type == EMLFitType.E_FIT_CLF:
            使用有监督学习分类self.estimator.clf
        self.fiter_type == EMLFitType.E_FIT_HMM:
            使用无监督学习hmm，self.estimator.hmm
        self.fiter_type == EMLFitType.E_FIT_PCA:
            使用无监督学习pca，self.estimator.pca
        self.fiter_type == EMLFitType.E_FIT_KMEAN:
            使用无监督学习kmean，self.estimator.kmean
        :return: 返回学习器对象
        """
        if self.fiter_type == EMLFitType.E_FIT_AUTO:
            if len(np.unique(self.y)) <= 10:
                # 小于等于10个class的y就认为是要用分类了
                fiter = self.estimator._classifier
            else:
                fiter = self.estimator._regressor
        elif self.fiter_type == EMLFitType.E_FIT_REG:
            fiter = self.estimator._regressor
        elif self.fiter_type == EMLFitType.E_FIT_CLF:
            fiter = self.estimator._classifier
        elif self.fiter_type == EMLFitType.E_FIT_HMM:
            if self.estimator._hmm is None:
                self.estimator.hmm_gaussian()
            fiter = self.estimator._hmm
        elif self.fiter_type == EMLFitType.E_FIT_PCA:
            if self.estimator._pca is None:
                self.estimator.pca_decomposition()
            fiter = self.estimator._pca
        elif self.fiter_type == EMLFitType.E_FIT_KMEAN:
            if self.estimator._kmean is None:
                self.estimator.kmean_cluster()
            fiter = self.estimator._kmean
        else:
            raise TypeError('self.fiter_type = {}, is error type'.format(
                self.fiter_type))

        return fiter

    @entry_wrapper(support=(EMLFitType.E_FIT_CLF, ))
    def cross_val_accuracy_score(self, cv=10, **kwargs):
        """
        被装饰器entry_wrapper(support=(EMLFitType.E_FIT_CLF,))装饰，
        即只支持有监督学习分类，使用cross_val_score对数据进行accuracy度量
        :param cv: 透传cross_val_score的参数，默认10
        :param kwargs: 外部可以传递x, y, 通过
                                x = kwargs.pop('x', self.x)
                                y = kwargs.pop('y', self.y)
                       确定传递self._do_cross_val_score中参数x，y，
                       装饰器使用的fiter_type，eg：ttn_abu.cross_val_accuracy_score(fiter_type=ml.EMLFitType.E_FIT_CLF)
        :return: cross_val_score返回的score序列，
                 eg: array([ 1.  ,  0.9 ,  1.  ,  0.9 ,  1.  ,  0.9 ,  1.  ,  0.9 ,  0.95,  1.  ])
        """
        x = kwargs.pop('x', self.x)
        y = kwargs.pop('y', self.y)
        return self._do_cross_val_score(x, y, cv,
                                        _EMLScoreType.E_SCORE_ACCURACY.value)

    @entry_wrapper(support=(EMLFitType.E_FIT_CLF, ))
    def cross_val_prob_accuracy_score(self, pb_threshold, cv=10, **kwargs):
        """
        被装饰器entry_wrapper(support=(EMLFitType.E_FIT_CLF,))装饰，
        即只支持有监督学习分类，拆分训练集，测试集，对所有数据进行一次predict_proba
        获取分类的概率（具体阅读ABuMLExecute.run_prob_cv_estimator），之后根据
        pb_threshold的值对结果概率进行二值转换，pb_threshold的均衡对影响数据的度量
        准确度等
                eg:
                    in: ttn_abu = AbuML.create_test_more_fiter()
                        ttn_abu.estimator.svc(probability=True)
                        ttn_abu.cross_val_prob_accuracy_score(pb_threshold=0.60)
                    out:
                        threshold=0.6 prob accuracy=0.83, effect cnt=870, effect rate=0.98, score=0.81
                    阀值0.6，准确率0.83，生效比例0.98，分数0.81

                    in:
                        ttn_abu.cross_val_prob_accuracy_score(pb_threshold=0.80)
                    out:
                        threshold=0.8 prob accuracy=0.87, effect cnt=718, effect rate=0.81, score=0.70
                    阀值0.8，准确率0.87 提高，生效比例0.81 降低，分数0.70 降低

                    in:
                        ttn_abu.cross_val_prob_accuracy_score(pb_threshold=0.85)
                    out:
                        threshold=0.85 prob accuracy=0.89, effect cnt=337, effect rate=0.38, score=0.34
                    阀值0.85，准确率0.89 再次提高，生效比例0.38 很低，分数0.34 降低

        即通过训练集数据寻找合适的prob值对数据的predict_prob进行非均衡处理，必然对交易的拦截进行非均衡
        处理，只对有很大概率的的交易进行拦截

        :param pb_threshold: binarize(y_prob, threshold=pb_threshold)中使用的二分阀值，float（0-1）
        :param cv: 透传ABuMLExecute.run_prob_cv_estimator中的cv参数，默认10，int
        :param show: 是否显示输出结果信息，默认显示
        :param kwargs: 外部可以传递x, y, 通过
                                x = kwargs.pop('x', self.x)
                                y = kwargs.pop('y', self.y)
                       确定传递self._do_cross_val_score中参数x，y
                       装饰器使用的fiter_type
        :return: accuracy, effect_cnt, effect_rate, score
        """
        if pb_threshold < 0.0 or pb_threshold > 1:
            kd_logger.info(
                'pb_threshold must > 0 and < 1! now={}'.format(pb_threshold))
            return

        x = kwargs.pop('x', self.x)
        y = kwargs.pop('y', self.y)
        fiter = self.get_fiter()

        y_prob = ml_execute.run_prob_cv_estimator(fiter, x, y, n_folds=cv)
        """
            eg: y_prob
            array([[ 0.8726,  0.1274],
                   [ 0.0925,  0.9075],
                   [ 0.2485,  0.7515],
                   ...,
                   [ 0.3881,  0.6119],
                   [ 0.7472,  0.2528],
                   [ 0.8555,  0.1445]])
        """
        y_prob_binarize = binarize(y_prob, threshold=pb_threshold)
        """
            根据参数中pb_threshold的值对y_prob进行二值化，二值化的结果中有一行全是1，或者全是0的，
            pb_threshold的值越是非均衡，结果中全是1，或者全是0的的数据越多，度量准确性的时候这些都
            是期权票
            eg: y_prob_binarize
            array([[ 1.,  0.],
                   [ 0.,  1.],
                   [ 1.,  1.],
                   ...,
                   [ 0.,  0.],
                   [ 1.,  0.],
                   [ 1.,  0.]])
        """
        # y_unique eg: array([ 0.,  1.])
        y_unique = np.unique(y)
        y_label_binarize_df = pd.get_dummies(y, prefix='true')
        """
            eg: y_label_binarize_df
                 true_0.0  true_1.0
            0           1         0
            1           0         1
            2           0         1
            3           0         1
            4           1         0
            5           1         0
            6           1         0
            7           1         0
            8           0         1
            9           0         1
            ..        ...       ...
            881         1         0
            882         1         0
            883         1         0
            884         1         0
            885         1         0
            886         1         0
            887         0         1
            888         1         0
            889         0         1
            890         1         0
        """
        y_prob_df = pd.DataFrame(
            y_prob_binarize,
            columns=['prob_{}'.format(y_label) for y_label in y_unique])
        """
            eg: y_prob_df
                 prob_0.0  prob_1.0
            0         1.0       0.0
            1         0.0       1.0
            2         0.0       1.0
            3         0.0       1.0
            4         1.0       0.0
            5         1.0       0.0
            6         1.0       0.0
            7         1.0       0.0
            8         0.0       0.0
            9         0.0       1.0
            ..        ...       ...
            881       1.0       0.0
            882       0.0       1.0
            883       1.0       0.0
            884       1.0       0.0
            885       1.0       0.0
            886       1.0       0.0
            887       0.0       1.0
            888       0.0       0.0
            889       1.0       0.0
            890       1.0       0.0
        """
        # 把两个df合并起来
        true_prob_df = pd.concat([y_label_binarize_df, y_prob_df], axis=1)
        """
            eg: true_prob_df
                 true_0.0  true_1.0  prob_0.0  prob_1.0
            0           1         0       1.0       0.0
            1           0         1       0.0       1.0
            2           0         1       0.0       1.0
            3           0         1       0.0       1.0
            4           1         0       1.0       0.0
            5           1         0       1.0       0.0
            6           1         0       1.0       0.0
            7           1         0       1.0       0.0
            8           0         1       0.0       0.0
            9           0         1       0.0       1.0
            ..        ...       ...       ...       ...
            881         1         0       1.0       0.0
            882         1         0       0.0       1.0
            883         1         0       1.0       0.0
            884         1         0       1.0       0.0
            885         1         0       1.0       0.0
            886         1         0       1.0       0.0
            887         0         1       0.0       1.0
            888         1         0       0.0       0.0
            889         0         1       1.0       0.0
            890         1         0       1.0       0.0
        """

        # 即筛选出非均衡阀值情况下有效的投票行index
        vote_index = (y_prob_df.sum(axis=1) > 0) & (y_prob_df.sum(axis=1) < 2)
        """
            需要过滤非均衡阀值情况下如pb_threshold = 0.1，都投了1
            和如pb_threshold = 0.9，都不进行投票（全是0）的情况
            eg: pb_threshold = 0.1
                 prob_0.0  prob_1.0
            0         1.0       1.0
            1         1.0       1.0
            2         1.0       1.0
            3         1.0       1.0
            4         1.0       1.0
            5         1.0       1.0
            ..        ...       ...
            eg: pb_threshold = 0.9
                 prob_0.0  prob_1.0
            0         0.0       0.0
            1         0.0       0.0
            2         0.0       0.0
            3         0.0       0.0
            4         0.0       0.0
            5         0.0       0.0
            ..        ...       ...
        """

        # 再次进行拆开，根据vote_index
        # noinspection PyUnresolvedReferences
        true_df = true_prob_df[vote_index].filter(regex='true*')
        # noinspection PyUnresolvedReferences
        prob_df = true_prob_df[vote_index].filter(regex='prob*')
        """
            prob_df即是y_prob_df中拥有有效投票的序列，true_df对应prob_df的index
            eg：prob_df
                 prob_0.0  prob_1.0
            0         1.0       0.0
            1         0.0       1.0
            2         0.0       1.0
            3         0.0       1.0
            4         1.0       0.0
            5         1.0       0.0
            6         1.0       0.0
            7         1.0       0.0
            9         0.0       1.0
            10        0.0       1.0
            ..        ...       ...
            880       0.0       1.0
            881       1.0       0.0
            882       0.0       1.0
            883       1.0       0.0
            884       1.0       0.0
            885       1.0       0.0
            886       1.0       0.0
            887       0.0       1.0
            889       1.0       0.0
            890       1.0       0.0
        """

        # 生效数量，投票不合格的不做准确率统计
        effect_cnt = prob_df.shape[0]
        # 生效率：effect_cnt / y.shape[0]
        effect_rate = effect_cnt / y.shape[0]
        # 生效的数据准确率
        accuracy = 0.0
        if effect_cnt > 0:
            accuracy = metrics.accuracy_score(true_df, prob_df)
        # 分数：生效比例 ＊ 生效准确率（0-1）
        score = effect_rate * accuracy
        kd_logger.info(
            'threshold={} prob accuracy={:.2f}, effect cnt={}, effect rate={:.2f}, score={:.2f}'
            .format(pb_threshold, accuracy, effect_cnt, effect_rate, score))
        return accuracy, effect_cnt, effect_rate, score

    @entry_wrapper(support=(EMLFitType.E_FIT_KMEAN, ))
    def cross_val_silhouette_score(self, cv=10, **kwargs):
        """
        被装饰器entry_wrapper(support=(EMLFitType.E_FIT_KMEAN, ))装饰，即
        只支持无监督学习kmean的cv验证，使用silhouette_score对聚类后的结果labels_
        进行度量使用silhouette_score
        :param cv: 透传run_silhouette_cv_estimator的参数，默认10
        :param kwargs: 外部可以传递x 通过
                                x = kwargs.pop('x', self.x)
                       确定传递ABuMLExecute.run_silhouette_cv_estimator中参数x
                       装饰器使用的fiter_type，
                            eg：ttn_abu.cross_val_silhouette_score(fiter_type=ml.EMLFitType.E_FIT_KMEAN)
        :return: run_silhouette_cv_estimator返回的score序列，
                 eg: array([ 0.6322,  0.6935,  0.7187,  0.6887,  0.6699,  0.703 ,  0.6922,
                            0.7049,  0.6462,  0.6755])
        """
        x = kwargs.pop('x', self.x)
        fiter = self.get_fiter()

        scores = ml_execute.run_silhouette_cv_estimator(fiter, x, n_folds=cv)
        scores = np.array(scores)
        kd_logger.info('{} score mean: {}'.format(fiter.__class__.__name__,
                                                  scores.mean()))
        return scores

    @entry_wrapper(support=(EMLFitType.E_FIT_CLF, EMLFitType.E_FIT_REG))
    def cross_val_mean_squared_score(self, cv=10, **kwargs):
        """
        被装饰器entry_wrapper(support=(EMLFitType.E_FIT_CLF, EMLFitType.E_FIT_REG))装饰，
        即支持有监督学习回归和分类，使用cross_val_score对数据进行rmse度量
        :param cv: 透传cross_val_score的参数，默认10
        :param kwargs: 外部可以传递x, y, 通过
                                x = kwargs.pop('x', self.x)
                                y = kwargs.pop('y', self.y)
                       确定传递self._do_cross_val_score中参数x，y，
                       以及装饰器使用的fiter_type，eg：ttn_abu.cross_val_roc_auc_score(fiter_type=ml.EMLFitType.E_FIT_CLF)
        :return: cross_val_score返回的score序列，
                 eg: array([-0.1889, -0.1667, -0.2135, -0.1348, -0.1573, -0.2022, -0.1798,
                            -0.2022, -0.1348, -0.1705])
        """
        x = kwargs.pop('x', self.x)
        y = kwargs.pop('y', self.y)
        return self._do_cross_val_score(x, y, cv,
                                        _EMLScoreType.E_SCORE_MSE.value)

    @entry_wrapper(support=(EMLFitType.E_FIT_CLF, ))
    def cross_val_roc_auc_score(self, cv=10, **kwargs):
        """
        被装饰器entry_wrapper(support=(EMLFitType.E_FIT_CLF,))装饰，
        即支持有监督学习分类，使用cross_val_score对数据进行roc_auc度量，如果数据的y的
        label标签 > 2，通过label_binarize将label标签进行二值化处理，
        依次计算二值化的列的roc_auc，结果返回score最好的数据度量
        :param cv: 透传cross_val_score的参数，默认10
        :param kwargs: 外部可以传递x, y, 通过
                                x = kwargs.pop('x', self.x)
                                y = kwargs.pop('y', self.y)
                       确定传递self._do_cross_val_score中参数x，y，
                       以及装饰器使用的fiter_type，eg：ttn_abu.cross_val_roc_auc_score(fiter_type=ml.EMLFitType.E_FIT_REG)
        :return: cross_val_score返回的score序列，
                 eg: array([ 1.  ,  0.9 ,  1.  ,  0.9 ,  1.  ,  0.9 ,  1.  ,  0.9 ,  0.95,  1.  ])
        """
        x = kwargs.pop('x', self.x)
        y = kwargs.pop('y', self.y)
        return self._do_cross_val_score(x, y, cv,
                                        _EMLScoreType.E_SCORE_ROC_AUC.value)

    @entry_wrapper()
    def feature_selection(self, **kwargs):
        """
            被装饰器entry_wrapper()装饰，默认参数即支持有监督和无监督学习
            监督学习通过sklern中的RFE包装拟合器进行fit(x, y)，对数据的特征进行ranking和support评定

            eg：
                RFE selection
                             ranking support
                SibSp              1    True
                Parch              1    True
                Cabin_No           1    True
                Cabin_Yes          7   False
                Embarked_C         2   False
                Embarked_Q         3   False
                Embarked_S         5   False
                Sex_female         8   False
                Sex_male           1    True
                Pclass_1           4   False
                Pclass_2           6   False
                Pclass_3           1    True
                Age_scaled         1    True
                Fare_scaled        1    True

            无监督学习通过sklern中的VarianceThreshold进行fit(x)，根据x的方差进行特征评定
            eg:
                unsupervised VarianceThreshold
                            support
                SibSp          True
                Parch          True
                Cabin_No       True
                Cabin_Yes      True
                Embarked_C     True
                Embarked_Q     True
                Embarked_S     True
                Sex_female     True
                Sex_male       True
                Pclass_1       True
                Pclass_2       True
                Pclass_3       True
                Age_scaled     True
                Fare_scaled    True

        :param show: 是否在内部输出打印结果
        :param kwargs: 外部可以传递x, y, 通过
                                x = kwargs.pop('x', self.x)
                                y = kwargs.pop('y', self.y)
                       确定传递self._do_cross_val_score中参数x，y，
                       以及装饰器使用的fiter_type，eg：ttn_abu.feature_selection(fiter_type=ml.EMLFitType.E_FIT_REG)
        """
        x = kwargs.pop('x', self.x)
        y = kwargs.pop('y', self.y)
        fiter = self.get_fiter()

        if self.is_supervised_learning():
            selector = RFE(fiter)
            selector.fit(x, y)
            feature_df = pd.DataFrame(
                {
                    'support': selector.support_,
                    'ranking': selector.ranking_
                },
                index=self.df.columns[1:])
            print('RFE selection')
            print(feature_df)
        else:
            selector = VarianceThreshold()
            selector.fit(x)
            feature_df = pd.DataFrame({'support': selector.get_support()},
                                      index=self.df.columns[1:])

            print('unsupervised VarianceThreshold')
            print(feature_df)
        return feature_df

    @entry_wrapper(support=(EMLFitType.E_FIT_CLF, EMLFitType.E_FIT_REG))
    def importances_coef_pd(self, **kwargs):
        """
        被装饰器entry_wrapper(support=(EMLFitType.E_FIT_CLF, EMLFitType.E_FIT_REG))装饰，
        即支持有监督学习回归和分类，根据fit后的feature_importances_或者coef_和原始self.df.columns
        组成pd.DataFrame对象返回
            eg：
                    feature	    importance
                7	Sex_female	0.0000
                10	Pclass_2	0.0018
                3	Cabin_Yes	0.0033
                5	Embarked_Q	0.0045
                9	Pclass_1	0.0048
                4	Embarked_C	0.0098
                6	Embarked_S	0.0105
                1	Parch	0.0154
                2	Cabin_No	0.0396
                0	SibSp	0.0506
                11	Pclass_3	0.0790
                13	Fare_scaled	0.1877
                12	Age_scaled	0.2870
                8	Sex_male	0.3060

                    coef	            columns
                0	[-0.344229036121]	SibSp
                1	[-0.1049314305]	Parch
                2	[0.0]	Cabin_No
                3	[0.902140498996]	Cabin_Yes
                4	[0.0]	Embarked_C
                5	[0.0]	Embarked_Q
                6	[-0.417254399259]	Embarked_S
                7	[1.95656682017]	Sex_female
                8	[-0.677432099492]	Sex_male
                9	[0.3411515052]	Pclass_1
                10	[0.0]	Pclass_2
                11	[-1.19413332987]	Pclass_3
                12	[-0.523782082975]	Age_scaled
                13	[0.0844326510536]	Fare_scaled
        :param kwargs: 外部可以传递x, y, 通过
                                x = kwargs.pop('x', self.x)
                                y = kwargs.pop('y', self.y)
                       以及装饰器使用的fiter_type，eg：ttn_abu.importances_coef_pd(fiter_type=ml.EMLFitType.E_FIT_REG)
        :return: pd.DataFrame对象
        """
        if not hasattr(self, 'df'):
            raise ValueError('please make a df func first!')
        x = kwargs.pop('x', self.x)
        y = kwargs.pop('y', self.y)

        fiter = self.get_fiter()
        # 训练前进行clone(fiter)
        fiter = clone(fiter)
        fiter.fit(x, y)

        # self.echo_info(fiter)
        if hasattr(fiter, 'feature_importances_'):
            return pd.DataFrame({
                'feature': list(self.df.columns)[1:],
                'importance': fiter.feature_importances_
            }).sort_values('importance')
        elif hasattr(fiter, 'coef_'):
            return pd.DataFrame({
                "columns": list(self.df.columns)[1:],
                "coef": list(fiter.coef_.T)
            })

        else:
            kd_logger.info('fiter not hasattr feature_importances_ or coef_!')

    @entry_wrapper(support=(EMLFitType.E_FIT_CLF, ))
    def train_test_split_xy(self, test_size=0.1, random_state=0, **kwargs):
        """
            被entry_wrapper(support=(EMLFitType.E_FIT_CLF, ))装饰，即只支持分类
            首先使用train_test_split将x，y根据参数test_size切割训练集和测试集，
            显示数据集，训练集，测试集的数量，针对训练集数据进行训练，使用训练好的
            分类器对测试集x进行predict，对结果分别使用metrics.accuracy_score,
            metrics.precision_score, metrics.recall_score度量准确率，查准率，
            和召回率，多label的的情况下使用average = 'macro'对precision_score和
            recall_score进行度量，最后显示分类结果混淆矩阵以及metrics.classification_report
            情况

            eg:
                x-y:(891, 14)-(891,)
                train_x-train_y:(801, 14)-(801,)
                test_x-test_y:(90, 14)-(90,)
                accuracy = 0.77
                precision_score = 0.74
                recall_score = 0.72
                          Predicted
                         |  0  |  1  |
                         |-----|-----|
                       0 |  41 |  10 |
                Actual   |-----|-----|
                       1 |  11 |  28 |
                         |-----|-----|
                             precision    recall  f1-score   support

                        0.0       0.79      0.80      0.80        51
                        1.0       0.74      0.72      0.73        39

                avg / total       0.77      0.77      0.77        90

        :param test_size: 测试集占比例，float，默认0.1，即将数据分10份，一份做为测试集
        :param random_state: 透传给train_test_split的随机参数
        :param kwargs: 外部可以传递x, y, 通过
                                x = kwargs.pop('x', self.x)
                                y = kwargs.pop('y', self.y)
                       以及装饰器使用的fiter_type，eg：ttn_abu.train_test_split_xy(fiter_type=ml.EMLFitType.E_FIT_CLF)
        """
        x = kwargs.pop('x', self.x)
        y = kwargs.pop('y', self.y)
        train_x, test_x, train_y, test_y = train_test_split(
            x, y, test_size=test_size, random_state=random_state)
        print('x-y:{}-{}'.format(x.shape, y.shape))
        print('train_x-train_y:{}-{}'.format(train_x.shape, train_y.shape))
        print('test_x-test_y:{}-{}'.format(test_x.shape, test_y.shape))

        fiter = self.get_fiter()
        # 训练前进行clone(fiter)
        fiter = clone(fiter)
        # 针对训练集数据进行训练
        clf = fiter.fit(train_x, train_y)
        # 使用训练好的分类器对测试集x进行predict，结果y_predict
        y_predict = clf.predict(test_x)

        # 度量分类准确率
        print("accuracy = %.2f" % (metrics.accuracy_score(test_y, y_predict)))
        # precision_score和predictions在二分类的情况下使用binary
        average = 'binary'
        if len(np.unique(y)) != 2:
            # “micro表示在多分类中的对所有label进行averaging计算平均precision，recall以及F值等度量
            average = 'macro'
        # 度量分类查准率
        print("precision_score = %.2f" %
              (metrics.precision_score(test_y, y_predict, average=average)))
        # 度量分类召回率
        print("recall_score = %.2f" %
              (metrics.recall_score(test_y, y_predict, average=average)))
        # 混淆矩阵以及metrics.classification_report
        self._confusion_matrix_with_report(test_y,
                                           y_predict,
                                           labels=np.unique(y))

    def train_test_split_df(self, test_size=0.1, random_state=0, **kwargs):
        """
        套接封装train_test_split_xy，外部传递pd.DataFrame参数时使用
        :param test_size: 透传参数train_test_split_xy
        :param random_state: 透传参数train_test_split_xy
        :param kwargs: 通过 df = kwargs.pop('df', self.df)弹出传递的pd.DataFrame对象进行x，y分解
                       y = matrix[:, 0]，即硬编码分类y在第一列，外部传递的df对象需要遵循
                       以及装饰器使用的fiter_type，eg：ttn_abu.train_test_split_df(fiter_type=ml.EMLFitType.E_FIT_CLF)
        """
        df = kwargs.pop('df', self.df)
        matrix = df.as_matrix()
        y = matrix[:, 0]
        x = matrix[:, 1:]
        self.train_test_split_xy(test_size=test_size,
                                 random_state=random_state,
                                 x=x,
                                 y=y,
                                 **kwargs)

    @entry_wrapper()
    def fit(self, **kwargs):
        """
        包装fit操作，根据是否是有监督学习来区别
        使用fit(x, y)还是fit(x)

            eg:
               in:   iris_abu.estimator.random_forest_classifier()
                     iris_abu.fit()
               out:
                     RandomForestClassifier(bootstrap=True, class_weight=None, criterion='gini',
                                            max_depth=None, max_features='auto', max_leaf_nodes=None,
                                            min_impurity_split=1e-07, min_samples_leaf=1,
                                            min_samples_split=2, min_weight_fraction_leaf=0.0,
                                            n_estimators=100, n_jobs=1, oob_score=False, random_state=None,
                                            verbose=0, warm_start=False)
        :param kwargs:
        :return: fit(x, y)或者fit(x)操作后返回
        """
        x = kwargs.pop('x', self.x)
        y = kwargs.pop('y', self.y)
        fiter = self.get_fiter()
        if hasattr(fiter, 'fit'):
            if self.is_supervised_learning():
                fit_self = fiter.fit(x, y)
            else:
                fit_self = fiter.fit(x)
            return fit_self
        else:
            kd_logger.info('{} not support fit'.format(fiter))

    @entry_wrapper()
    def fit_transform(self, **kwargs):
        """
        被装饰器@entry_wrapper()装饰，默认参数即支持有监督和无监督学习，
        内部通过检测isinstance(fiter, TransformerMixin) or hasattr(fiter, 'fit_transform')
        来判定是否可以fit_transform

        eg：
            input:  ttn_abu.x.shape
            output: (891, 14)

            input:  ttn_abu.fit_transform(fiter_type=ml.EMLFitType.E_FIT_PCA).shape
            output: (891, 4)

            input:  ttn_abu.fit_transform(fiter_type=ml.EMLFitType.E_FIT_KMEAN).shape
            output: (891, 2)

        :param kwargs: 外部可以传递x, y, 通过
                                x = kwargs.pop('x', self.x)
                                y = kwargs.pop('y', self.y)
                       以及装饰器使用的fiter_type，eg：ttn_abu.fit_transform(fiter_type=ml.EMLFitType.E_FIT_CLF)
        :return: fit_transform后的转换结果矩阵
        """
        fiter = self.get_fiter()
        if isinstance(fiter, TransformerMixin) or hasattr(
                fiter, 'fit_transform'):
            x = kwargs.pop('x', self.x)
            y = kwargs.pop('y', self.y)
            if self.is_supervised_learning():
                trans = fiter.fit_transform(x, y)
            else:
                trans = fiter.fit_transform(x)
            return trans
        else:
            kd_logger.info('{} not support fit_transform'.format(fiter))

    def predict(self, x, pre_fit=True, **kwargs):
        """
        call有监督和无监督学习的predict，套接self._predict_callback方法，其
        被装饰器params_to_numpy装饰，将predict参数中所有可迭代序列变成numpy，
        即统一将x转换numpy序列,
            eg:
                test_input = pd.DataFrame.from_dict({'x0': 6.8, 'x1': 3.5,  'x2': 5.4, 'x3': 2.2}, orient='index')
                test_input:
                                0
                            x0	6.8
                            x1	3.5
                            x2	5.4
                            x3	2.2

                iris_abu.predict(test_input)

                params_to_numpy装饰后参数中的x转换为：array([[ 6.8,  3.5,  5.4,  2.2]]) numpy array

        被装饰器entry_wrapper()装饰，默认参数即支持有监督和无监督学习
        :param x: 可迭代序列，通过params_to_numpy装饰统一转换numpy
        :param pre_fit: bool，默认pre_fit True, 代表每次predict前先fit，
                        如在乎效率的情况下，需要在外部先fit后，参数pre_fit置False
        :param kwargs: 装饰器使用的fiter_type，eg：ttn_abu.predict(test_input, fiter_type=ml.EMLFitType.E_FIT_KMEAN)
        :return: eg: array([2])
        """
        return self._predict_callback('predict', x, pre_fit=pre_fit, **kwargs)

    def predict_proba(self, x, pre_fit=True, **kwargs):
        """
        call有监督和无监督学习的predict_proba，套接self._predict_callback方法，其
        被装饰器params_to_numpy装饰，将predict参数中所有可迭代序列变成numpy，
        即统一将x转换numpy序列,
            eg:
                test_input = pd.DataFrame.from_dict({'x0': 6.8, 'x1': 3.5,  'x2': 5.4, 'x3': 2.2}, orient='index')
                test_input:
                                0
                            x0	6.8
                            x1	3.5
                            x2	5.4
                            x3	2.2

                iris_abu.predict_proba(test_input, fiter_type=ml.EMLFitType.E_FIT_CLF)

                params_to_numpy装饰后参数中的x转换为：array([[ 6.8,  3.5,  5.4,  2.2]]) numpy array

        被装饰器entry_wrapper()装饰，默认参数即支持有监督和无监督学习
        :param x: 通过params_to_numpy装饰统一转换numpy
        :param pre_fit: bool，默认pre_fit True, 代表每次predict前先fit，
                        如在乎效率的情况下，需要在外部先fit后，参数pre_fit置False
        :param kwargs: 装饰器使用的fiter_type，eg：iris_abu.predict_proba(test_input, fiter_type=ml.EMLFitType.E_FIT_CLF)
        :return: eg: array([[ 0.2,  0.3,  0.5]])
        """
        return self._predict_callback('predict_proba',
                                      x,
                                      pre_fit=pre_fit,
                                      **kwargs)

    def predict_proba_threshold(self,
                                x,
                                threshold,
                                default_ret,
                                pre_fit=True,
                                **kwargs):
        """
        eg：
            in:  ttn_abu = AbuML.create_test_more_fiter()
            in:  ttn_abu.estimator.svc(probability=True)
            in： ttn_abu.search_match_pos_threshold(0.85, 0.80, fiter_type=ml.EMLFitType.E_FIT_CLF)
            out: 0.770 satisfy require, accuracy:0.850, effect_rate:0.854

            即通过search_match_pos_threshold获取了predict_proba的二分阀值为0.77时，分类的正确率达到0.85， 且覆盖80%样本

            in:  test = np.array([ 1.    ,  0.    ,  0.    ,  1.    ,  1.    ,  0.    ,  0.    ,1.    ,  0.    ,  1.
                                ,  0.    ,  0.    ,  0.8132,  0.5868])
            in:  ttn_abu.predict_proba(test)
            out: array([[ 0.106,  0.894]])
            使用predict_proba得到的是分类的概率

            in:  ttn_abu.predict_proba_threshold(test, threshold=0.77 ,default_ret=0)
            out: 1
            使用predict_proba_threshold将阀值threshold设定0.77后，可以得到输出为1，即概率 0.894 > 0.77, 标签1分类成立

            in:  test2 = np.array([ 0.    ,  1.    ,  1.    ,  0.    ,  1.    ,  1.    ,  0.    ,1.    ,  0.    ,  0.
                                ,  0.    ,  1.    ,  0.7832,  0.2868])
            in:  ttn_abu.predict_proba(test2)
            out: array([[ 0.2372,  0.7628]])

            in:  ttn_abu.predict_proba_threshold(test2, threshold=0.77 ,default_ret=0)
            out: 0
            对test2做predict_proba_threshold返回0，因为0.7628 < 0.77, 标签1的分类不成立，返回default_ret

            应用场景：比如对交易进行拦截，实行高放行率，低拦截率，0代表放行，1代表拦截，
                    上述predict_proba_threshold(test2, threshold=0.77 ,default_ret=0)
                    即可实行对较大概率的交易进行拦截，即把握大的进行拦截，把握不大的默认选择放行
        :param x: 在predict_proba中通过params_to_numpy装饰统一转换numpy
        :param threshold: 对predict_proba结果进行二值化的阀值 eg： threshold=0.77
        :param default_ret: 使用predict_proba返回的矢量和不等于1时，即没有进行有效投票时返回的值：
                            eg：
                                    proba = np.where(proba >= threshold, 1, 0)
                                    if proba.sum() != 1:
                                        # eg: proba = array([[ 0.2328,  0.7672]])->array([[0, 0]])
                                        return default_ret
        :param pre_fit: bool，默认pre_fit True, 代表每次predict前先fit，
                        如在乎效率的情况下，需要在外部先fit后，参数pre_fit置False
        :param kwargs: 装饰器使用的fiter_type，
                        eg：iris_abu.predict_proba_threshold(test_input, , threshold=0.77 ,default_ret=0
                                                            fiter_type=ml.EMLFitType.E_FIT_CLF)
        :return: int，default_ret or proba.argmax()
        """
        # 套接self.predict_proba对x所描述的特征进行概率proba
        proba = self.predict_proba(x, pre_fit=pre_fit, **kwargs)
        # eg：array([[ 0.1063,  0.8937]]) -> array([[0, 1]])
        # noinspection PyTypeChecker
        proba = np.where(proba >= threshold, 1, 0)
        if proba.sum() != 1:
            # eg: proba = array([[ 0.2328,  0.7672]])->array([[0, 0]])
            return default_ret
        # 唯一最大值就是序列值为1的，通过argmax获取index，即y label
        return proba.argmax()

    @params_to_numpy
    @entry_wrapper()
    def _predict_callback(self, callback, x, pre_fit=True, **kwargs):
        """
        统一封装predict和predict需要的流程，使用callback做为具体实现
        :param callback: str字符类型，不是callable类型
        :param x: 可迭代序列，通过params_to_numpy装饰统一转换numpy
        :param pre_fit: bool，默认pre_fit True, 代表每次predict前先fit，
                        如在乎效率的情况下，需要在外部先fit后，参数pre_fit置False
        :param kwargs: 装饰器使用的fiter_type，eg：ttn_abu.predict(test_input, fiter_type=ml.EMLFitType.E_FIT_KMEAN)
        :return: eg: array([2])
        """
        # 标准化输入x
        x = x.reshape(1, -1)
        if self.x[0].reshape(1, -1).shape != x.shape:
            # predict中有check_input=True也check，返回训练集中的一个数据做为input x的示例
            kd_logger.info('input x must similar with {}'.format(self.x[0]))
            return

        if pre_fit:
            # 默认pre_fit True, 代表每次predict前先fit，如在乎效率的情况下，需要在外部先fit后，参数pre_fit置False
            self.fit(**kwargs)

        fiter = self.get_fiter()
        if not isinstance(callback, six.string_types):
            # callback必须是字符串类型
            kd_logger.info('callback must str, not {}'.format(type(callback)))
            return

        if hasattr(fiter, callback):
            if 'check_input' in list(
                    signature(fiter.predict).parameters.keys()):
                # 针对有check_input参数的，check_input True, 因为前面X.dtype np.float32等格式化问题
                return getattr(fiter, callback)(x, check_input=True)
            else:
                return getattr(fiter, callback)(x)
        else:
            kd_logger.info('{} not support {}'.format(fiter, callback))

    def search_match_neg_threshold(self,
                                   accuracy_match=0,
                                   effect_rate_match=0,
                                   neg_num=50,
                                   **kwargs):
        """
        套接self.cross_val_prob_accuracy_score，通过np.linspace(0.01, 0.50, num=neg_num)[::-1]生成
        pb_threshold参数序列，这里linspace的start从0.01至0.50后[::-1]倒序，依次迭代生成的阀值参数
        ，当cross_val_prob_accuracy_score返回的正确率大于 参数中accuracy_match且返回的生效率大于参数中的effect_rate_match，
        匹配寻找成功，中断迭代操作，返回寻找到的满足条件的阀值，返回的阀值应用场景阅读predict_proba_threshold函数

            eg:
                in:  ttn_abu.search_match_neg_threshold(0.85, 0.80, fiter_type=ml.EMLFitType.E_FIT_CLF)
                out: 0.220 satisfy require, accuracy:0.852, effect_rate:0.844

        :param accuracy_match: 寻找阀值条件，需要当cross_val_prob_accuracy_score返回的正确率大于accuracy_match，
                               float， 范围（0-1），默认值0
        :param effect_rate_match: 寻找阀值条件，需要当cross_val_prob_accuracy_score返回的生效率大于effect_rate_match，
                               float， 范围（0-1），默认值0
        :param neg_num: 透传neg_thresholds = np.linspace(0.01, 0.50, num=neg_num)[::-1]的参数，默认50
        :param kwargs: 装饰器使用的fiter_type，
                    eg: iris_abu.search_match_neg_threshold(0.85, 0.80, fiter_type=ml.EMLFitType.E_FIT_CLF)
        :return: 返回寻找到的满足条件的阀值，float
        """
        neg_thresholds = np.linspace(0.01, 0.50, num=neg_num)[::-1]
        """
            eg: neg_thresholds
                array([ 0.5 ,  0.49,  0.48,  0.47,  0.46,  0.45,  0.44,  0.43,  0.42,
            0.41,  0.4 ,  0.39,  0.38,  0.37,  0.36,  0.35,  0.34,  0.33,
            0.32,  0.31,  0.3 ,  0.29,  0.28,  0.27,  0.26,  0.25,  0.24,
            0.23,  0.22,  0.21,  0.2 ,  0.19,  0.18,  0.17,  0.16,  0.15,
            0.14,  0.13,  0.12,  0.11,  0.1 ,  0.09,  0.08,  0.07,  0.06,
            0.05,  0.04,  0.03,  0.02,  0.01])
        """
        with Progress(len(neg_thresholds), 0,
                      'search neg threshold') as search_neg_progress:
            for neg in neg_thresholds:
                accuracy, _, effect_rate, _ = self.cross_val_prob_accuracy_score(
                    neg, show=False, **kwargs)
                search_neg_progress.show(
                    ext='threshold:{:.2f} accuracy:{:.2f}, effect_rate:{:.2f}'.
                    format(neg, accuracy, effect_rate))
                if accuracy >= accuracy_match and effect_rate >= effect_rate_match:
                    # eg: 0.500 satisfy require, accuracy:0.940, effect_rate:1.000
                    kd_logger.info(
                        '{:.3f} satisfy require, accuracy:{:.3f}, effect_rate:{:.3f}'
                        .format(neg, accuracy, effect_rate))
                    # 返回寻找到的满足条件的阀值
                    return neg
        # 迭代完成所有neg_thresholds，没有找到符合参数需求的二分阀值
        kd_logger.info('neg_thresholds no satisfy require, search failed!')

    def search_match_pos_threshold(self,
                                   accuracy_match=0,
                                   effect_rate_match=0,
                                   pos_num=50,
                                   **kwargs):
        """
        套接self.cross_val_prob_accuracy_score，通过np.linspace(0.50, 0.99, num=neg_num)生成
        pb_threshold参数序列，这里linspace的start从0.50至0.99正序，依次迭代生成的阀值参数
        ，当cross_val_prob_accuracy_score返回的正确率大于 参数中accuracy_match且返回的生效率大于参数中的effect_rate_match，
        匹配寻找成功，中断迭代操作，返回寻找到的满足条件的阀值，返回的阀值应用场景阅读predict_proba_threshold函数

                eg:
                    in: ttn_abu.search_match_pos_threshold(0.85, 0.80, fiter_type=ml.EMLFitType.E_FIT_CLF)
                    out: 0.770 satisfy require, accuracy:0.850, effect_rate:0.854

        :param accuracy_match: 寻找阀值条件，需要当cross_val_prob_accuracy_score返回的正确率大于accuracy_match，
                               float， 范围（0-1），默认值0
        :param effect_rate_match: 寻找阀值条件，需要当cross_val_prob_accuracy_score返回的生效率大于effect_rate_match，
                               float， 范围（0-1），默认值0
        :param pos_num: 透传neg_thresholds = np.linspace(0.50, 0.99, num=neg_num)的参数，默认50
        :param kwargs: 装饰器使用的fiter_type，
                    eg: iris_abu.search_match_pos_threshold(0.85, 0.80, fiter_type=ml.EMLFitType.E_FIT_CLF)
        :return: 返回寻找到的满足条件的阀值，float
        """
        pos_thresholds = np.linspace(0.50, 0.99, num=pos_num)
        """
            eg: array([ 0.5 ,  0.51,  0.52,  0.53,  0.54,  0.55,  0.56,  0.57,  0.58,
                0.59,  0.6 ,  0.61,  0.62,  0.63,  0.64,  0.65,  0.66,  0.67,
                0.68,  0.69,  0.7 ,  0.71,  0.72,  0.73,  0.74,  0.75,  0.76,
                0.77,  0.78,  0.79,  0.8 ,  0.81,  0.82,  0.83,  0.84,  0.85,
                0.86,  0.87,  0.88,  0.89,  0.9 ,  0.91,  0.92,  0.93,  0.94,
                0.95,  0.96,  0.97,  0.98,  0.99])
        """
        with Progress(len(pos_thresholds), 0,
                      'search pos threshold') as search_pos_progress:
            for neg in pos_thresholds:
                accuracy, _, effect_rate, _ = self.cross_val_prob_accuracy_score(
                    neg, show=False, **kwargs)
                search_pos_progress.show(
                    ext='threshold:{:.2f} accuracy:{:.2f}, effect_rate:{:.2f}'.
                    format(neg, accuracy, effect_rate))
                if accuracy >= accuracy_match and effect_rate >= effect_rate_match:
                    # eg: 0.500 satisfy require, accuracy:0.940, effect_rate:1.000
                    kd_logger.info(
                        '{:.3f} satisfy require, accuracy:{:.3f}, effect_rate:{:.3f}'
                        .format(neg, accuracy, effect_rate))
                    # 返回寻找到的满足条件的阀值
                    return neg
        # 迭代完成所有pos_thresholds，没有找到符合参数需求的二分阀值
        kd_logger.info('pos_thresholds no satisfy require, search failed!')

    def bagging_classifier_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'max_samples': np.arange(1, 5), 'n_estimators': np.arange(100, 300, 50)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的BaggingClassifier对象
        """

        return self.estimator.bagging_classifier_best(self.x, self.y, **kwargs)

    def bagging_regressor_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'max_samples': np.arange(1, 5), 'n_estimators': np.arange(100, 300, 50)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的BaggingRegressor对象
        """
        return self.estimator.bagging_regressor_best(self.x, self.y, **kwargs)

    def adaboost_regressor_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'learning_rate': np.arange(0.2, 1.2, 0.2),
                                         'n_estimators': np.arange(10, 100, 10)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的AdaBoostRegressor对象
        """
        return self.estimator.adaboost_regressor_best(self.x, self.y, **kwargs)

    def adaboost_classifier_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'learning_rate': np.arange(0.2, 1.2, 0.2),
                                          'n_estimators': np.arange(10, 100, 10)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的AdaBoostClassifier对象
        """
        return self.estimator.adaboost_classifier_best(self.x, self.y,
                                                       **kwargs)

    def random_forest_classifier_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                         'n_estimators': np.arange(10, 150, 15)}

                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的RandomForestClassifier对象
        """
        return self.estimator.random_forest_classifier_best(
            self.x, self.y, **kwargs)

    def random_forest_regressor_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                         'n_estimators': np.arange(10, 150, 15)}

                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的RandomForestRegressor对象
        """
        return self.estimator.random_forest_regressor_best(
            self.x, self.y, **kwargs)

    def gradient_boosting_classifier_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'learning_rate': np.arange(0.1, 0.5, 0.05),
                                          'n_estimators': np.arange(50, 200, 10)}

                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的GradientBoostingClassifier对象
        """
        return self.estimator.gradient_boosting_classifier_best(
            self.x, self.y, **kwargs)

    def gradient_boosting_regressor_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'learning_rate': np.arange(0.1, 0.5, 0.05),
                                          'n_estimators': np.arange(10, 100, 10)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的GradientBoostingRegressor对象
        """
        return self.estimator.gradient_boosting_regressor_best(
            self.x, self.y, **kwargs)

    def lasso_regressor_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                         'max_depth': np.arange(1, 10, 1)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的Lasso对象
        """
        return self.estimator.lasso_regressor_best(self.x, self.y, **kwargs)

    def sgd_regressor_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                         'max_depth': np.arange(1, 10, 1)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的SGDRegressor对象
        """
        return self.estimator.sgd_regressor_best(self.x, self.y, **kwargs)

    def sgd_classifier_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                         'max_depth': np.arange(1, 10, 1)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的SGDClassifier对象
        """
        return self.estimator.sgd_classifier_best(self.x, self.y, **kwargs)

    def extra_tree_regressor_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                         'max_depth': np.arange(1, 10, 1)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的ExtraTreesRegressor对象
        """
        return self.estimator.extra_tree_regressor_best(
            self.x, self.y, **kwargs)

    def extra_tree_classifier_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                         'max_depth': np.arange(1, 10, 1)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的ExtraTreesClassifier对象
        """
        return self.estimator.extra_tree_classifier_best(
            self.x, self.y, **kwargs)

    def decision_tree_regressor_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                          'max_depth': np.arange(1, 10, 1)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的DecisionTreeRegressor对象
        """
        return self.estimator.decision_tree_regressor_best(
            self.x, self.y, **kwargs)

    def decision_tree_classifier_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                         'max_depth': np.arange(1, 10, 1)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的DecisionTreeClassifier对象
        """
        return self.estimator.decision_tree_classifier_best(
            self.x, self.y, **kwargs)

    def knn_classifier_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'algorithm': ['ball_tree', 'kd_tree', 'brute'],
                                         'n_neighbors': np.arange(1, 26, 1)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的KNeighborsClassifier对象
        """
        return self.estimator.knn_classifier_best(self.x, self.y, **kwargs)

    def xgboost_regressor_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'algorithm': ['ball_tree', 'kd_tree', 'brute'],
                                         'n_neighbors': np.arange(1, 26, 1)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的XGBRegressor对象
        """
        return self.estimator.xgboost_regressor_best(self.x, self.y, **kwargs)

    def xgboost_classifier_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'algorithm': ['ball_tree', 'kd_tree', 'brute'],
                                         'n_neighbors': np.arange(1, 26, 1)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的XGBClassifier对象
        """
        return self.estimator.xgboost_classifier_best(self.x, self.y, **kwargs)

    def lgbm_regressor_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'algorithm': ['ball_tree', 'kd_tree', 'brute'],
                                         'n_neighbors': np.arange(1, 26, 1)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的LGBMRegressor对象
        """
        return self.estimator.lgbm_regressor_best(self.x, self.y, **kwargs)

    def lgbm_classifier_best(self, **kwargs):
        """
        :param kwargs: 关键字可选参数param_grid: 最优字典关键字参数
                        eg：param_grid = {'algorithm': ['ball_tree', 'kd_tree', 'brute'],
                                         'n_neighbors': np.arange(1, 26, 1)}
                       关键字可选参数assign: 是否保存实例化后最优参数的学习器对象，默认True
                       关键字可选参数n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
                       关键字可选参数show: 是否可视化最优参数搜索结果
        :return: 通过最优参数构造的LGBMClassifier对象
        """
        return self.estimator.lgbm_classifier_best(self.x, self.y, **kwargs)

    def _confusion_matrix_with_report(self, y_true, y_predict, labels=None):
        """
        封装metrics.confusion_matrix与metrics.classification_report，对分类
        结果进行度量
        :param y_true: 原始数据中的y序列值
        :param y_predict: 使用分类器predict的y序列值
        :param labels: y序列中的lable序列
        """
        confusion_matrix = metrics.confusion_matrix(y_true,
                                                    y_predict,
                                                    labels=labels)
        if len(labels) == 2:
            # 只有二分问题Actual/Predicted
            print("          Predicted")
            print("         |  0  |  1  |")
            print("         |-----|-----|")
            print("       0 | %3d | %3d |" %
                  (confusion_matrix[0, 0], confusion_matrix[0, 1]))
            print("Actual   |-----|-----|")
            print("       1 | %3d | %3d |" %
                  (confusion_matrix[1, 0], confusion_matrix[1, 1]))
            print("         |-----|-----|")
        else:
            # 非二分问题直接显示confusion_matrix
            kd_logger.info("Confusion Matrix: \n{}".format(confusion_matrix))

        print(metrics.classification_report(y_true, y_predict))

    def _decomposition_2x(self, x, use_pca):
        """
        通过pca进行降维或者选中两个importances最大的特征将
        x变成只有两个维度的特征矩阵
        :param x: 进行降维的特征矩阵
        :param use_pca: 是否使用pca进行特征降维
        :return: 降维后的特征矩阵，降维后的矩阵只有两个维度
        """
        if use_pca:
            # 构造一个临时的pca进行降维，n_components=2只保留两个维度
            pca_2n = self.estimator.pca_decomposition(n_components=2,
                                                      assign=False)
            x = pca_2n.fit_transform(x)
            """
                eg:
                x before fit_transform:
                array([[ 1.    ,  0.    ,  1.    , ...,  1.    , -0.5614, -0.5024],
                       [ 1.    ,  0.    ,  0.    , ...,  0.    ,  0.6132,  0.7868],
                       [ 0.    ,  0.    ,  1.    , ...,  1.    , -0.2677, -0.4889],
                       ...,
                       [ 1.    ,  2.    ,  1.    , ...,  1.    , -0.9924, -0.1763],
                       [ 0.    ,  0.    ,  0.    , ...,  0.    , -0.2677, -0.0444],
                       [ 0.    ,  0.    ,  1.    , ...,  1.    ,  0.1727, -0.4924]])
                x fit_transform:
                array([[ 0.3805, -1.0005],
                       [ 0.0586,  1.7903],
                       [-0.3162, -0.7404],
                       ...,
                       [ 1.6132, -0.5185],
                       [-0.5952,  0.6252],
                       [-0.7428, -0.7119]])
            """
        else:
            # 选中两个importances最大的两个去画决策边界
            importances = self.importances_coef_pd()
            """
                        feature  importance
                0         SibSp      0.0000
                1         Parch      0.0000
                3     Cabin_Yes      0.0000
                4    Embarked_C      0.0000
                5    Embarked_Q      0.0000
                6    Embarked_S      0.0000
                7    Sex_female      0.0000
                9      Pclass_1      0.0000
                10     Pclass_2      0.0000
                12   Age_scaled      0.0000
                13  Fare_scaled      0.0000
                2      Cabin_No      0.0831
                11     Pclass_3      0.1836
                8      Sex_male      0.7333
            """
            if importances is None:
                kd_logger.info(
                    'self.importances_coef_pd() importances is None!!!')
                return
            # 根据importance排序特征重要程度，拿出最重要的两个维度index序列, eg: most_two=[8, 11]
            most_two = sorted(
                importances.sort_values('importance').index[-2:].tolist())
            # 从x中根据most_two拼接出一个新的x矩阵，只有两个维度
            x = np.concatenate((x[:, most_two[0]][:, np.newaxis],
                                x[:, most_two[1]][:, np.newaxis]),
                               axis=1)

        return x

    def _do_cross_val_score(self, x, y, cv, scoring):
        """
        封装sklearn中cross_val_score方法， 参数x, y, cv, scoring透传cross_val_score
        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param cv: 透传cross_val_score，cv参数，int
        :param scoring: 透传cross_val_score, 使用的度量方法
        :return: cross_val_score返回的score序列，
                 eg: array([ 1.  ,  0.9 ,  1.  ,  0.9 ,  1.  ,  0.9 ,  1.  ,  0.9 ,  0.95,  1.  ])
        """
        fiter = self.get_fiter()
        """
            eg: fiter
            DecisionTreeClassifier(class_weight=None, criterion='gini', max_depth=None,
                                    max_features=None, max_leaf_nodes=None,
                                    min_impurity_split=1e-07, min_samples_leaf=1,
                                    min_samples_split=2, min_weight_fraction_leaf=0.0,
                                    presort=False, random_state=None, splitter='best')
        """
        if scoring == _EMLScoreType.E_SCORE_ROC_AUC.value and len(
                np.unique(y)) != 2:
            # roc auc的度量下且y的label数量不是2项分类，首先使用label_binarize进行处理
            y_label_binarize = label_binarize(y, classes=np.unique(y))
            """
                eg：
                    np.unique(y) ＝ array([0, 1, 2])
                    y_label_binarize:
                    array([[1, 0, 0],
                           [1, 0, 0],
                           [1, 0, 0],
                           [1, 0, 0],
                           [1, 0, 0],
                           .........
                           [0, 1, 0],
                           [0, 1, 0],
                           [0, 1, 0],
                           [0, 1, 0],
                           [0, 1, 0],
                           [0, 1, 0],
                           [0, 1, 0],
                           [0, 1, 0],
                           .........
                           [0, 0, 1],
                           [0, 0, 1],
                           [0, 0, 1],
                           [0, 0, 1],
                           [0, 0, 1],
                           [0, 0, 1]])
            """
            label_cnt = len(np.unique(y))
            # one vs rest的score平均值的和
            mean_sum = 0
            # one vs rest中的最好score平均值
            best_mean = 0
            # 最好score平均值(best_mean)的score序列，做为结果返回
            scores = list()
            for ind in np.arange(0, label_cnt):
                # 开始 one vs rest
                _y = y_label_binarize[:, ind]
                """
                    eg: _y
                    array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                           1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                           1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
                """
                tmp_scores = cross_val_score(fiter,
                                             x,
                                             _y,
                                             cv=cv,
                                             scoring=scoring)
                tmp_mean = np.mean(tmp_scores)
                # one vs rest的score平均值进行叠加sum
                mean_sum += tmp_mean
                if len(scores) == 0 or tmp_mean > best_mean:
                    scores = tmp_scores
            # one vs rest的score平均值的和 / label_cnt
            mean_sc = mean_sum / label_cnt
        else:
            scores = cross_val_score(fiter, x, y, cv=cv, scoring=scoring)
            # 计算度量的score平均值，做为log输出，结果返回的仍然是scores
            mean_sc = -np.mean(np.sqrt(-scores)) if scoring == mean_squared_error_scorer \
                else np.mean(scores)
        kd_logger.info('{} score mean: {}'.format(fiter.__class__.__name__,
                                                  mean_sc))

        return scores


# -*- encoding:utf-8 -*-
"""封装常用的分析方式及流程模块"""

import numpy as np
from scipy import interp
from sklearn import metrics
from sklearn import tree
from sklearn.base import ClusterMixin, clone
from sklearn.metrics import roc_curve, auc
from ultron.ump.core.fixes import KFold

__all__ = [
    'run_silhouette_cv_estimator', 'run_prob_cv_estimator', 'run_cv_estimator',
    'plot_learning_curve', 'plot_decision_boundary', 'plot_confusion_matrices',
    'plot_roc_estimator', 'graphviz_tree', 'visualize_tree'
]


# noinspection PyUnresolvedReferences
def run_silhouette_cv_estimator(estimator, x, n_folds=10):
    """
    只针对kmean的cv验证，使用silhouette_score对聚类后的结果labels_
    进行度量使用silhouette_score，kmean的cv验证只是简单的通过np.random.choice
    进行随机筛选x数据进行聚类的silhouette_score度量，并不涉及训练集测试集
    :param estimator: keman或者支持estimator.labels_, 只通过if not isinstance(estimator, ClusterMixin)进行过滤
    :param x: x特征矩阵
    :param n_folds: int，透传KFold参数，切割训练集测试集参数，默认10
    :return: eg: array([ 0.693 ,  0.652 ,  0.6845,  0.6696,  0.6732,  0.6874,  0.668 ,
                         0.6743,  0.6748,  0.671 ])
    """

    if not isinstance(estimator, ClusterMixin):
        print('estimator must be ClusterMixin')
        return

    silhouette_list = list()
    # eg: n_folds = 10, len(x) = 150 -> 150 * 0.9 = 135
    choice_cnt = int(len(x) * ((n_folds - 1) / n_folds))
    choice_source = np.arange(0, x.shape[0])

    # 所有执行fit的操作使用clone一个新的
    estimator = clone(estimator)
    for _ in np.arange(0, n_folds):
        # 只是简单的通过np.random.choice进行随机筛选x数据
        choice_index = np.random.choice(choice_source, choice_cnt)
        x_choice = x[choice_index]
        estimator.fit(x_choice)
        # 进行聚类的silhouette_score度量
        silhouette_score = metrics.silhouette_score(x_choice,
                                                    estimator.labels_,
                                                    metric='euclidean')
        silhouette_list.append(silhouette_score)
    return silhouette_list


def run_prob_cv_estimator(estimator, x, y, n_folds=10):
    """
    通过KFold和参数n_folds拆分训练集和测试集，使用
    np.zeros((len(y), len(np.unique(y))))初始化prob矩阵，
    通过训练estimator.fit(x_train, y_train)后的分类器使用
    predict_proba将y_prob中的对应填数据

    :param estimator: 支持predict_proba的有监督学习, 只通过hasattr(estimator, 'predict_proba')进行过滤
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param n_folds: int，透传KFold参数，切割训练集测试集参数，默认10
    :return: eg: y_prob
                array([[ 0.8726,  0.1274],
                       [ 0.0925,  0.9075],
                       [ 0.2485,  0.7515],
                       ...,
                       [ 0.3881,  0.6119],
                       [ 0.7472,  0.2528],
                       [ 0.8555,  0.1445]])

    """
    if not hasattr(estimator, 'predict_proba'):
        print('estimator must has predict_proba')
        return

    # 所有执行fit的操作使用clone一个新的
    estimator = clone(estimator)
    kf = KFold(len(y), n_folds=n_folds, shuffle=True)
    y_prob = np.zeros((len(y), len(np.unique(y))))
    """
        根据y序列的数量以及y的label数量构造全是0的矩阵
        eg: y_prob
        array([[ 0.,  0.,  0.],
               [ 0.,  0.,  0.],
               [ 0.,  0.,  0.],
                ..............
               [ 0.,  0.,  0.],
               [ 0.,  0.,  0.],
               [ 0.,  0.,  0.],
    """

    for train_index, test_index in kf:
        x_train, x_test = x[train_index], x[test_index]
        y_train = y[train_index]

        # clf = clone(estimator)
        estimator.fit(x_train, y_train)
        # 使用predict_proba将y_prob中的对应填数据
        y_prob[test_index] = estimator.predict_proba(x_test)

    return y_prob


def run_cv_estimator(estimator, x, y, n_folds=10):
    """
    通过KFold和参数n_folds拆分训练集和测试集，使用
    y.copy()初始化y_pred矩阵，迭代切割好的训练集与测试集，
    不断通过 estimator.predict(x_test)将y_pred中的值逐步替换

    :param estimator: 有监督学习器对象
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param n_folds: int，透传KFold参数，切割训练集测试集参数，默认10
    :return: y_pred序列
    """
    if not hasattr(estimator, 'predict'):
        print('estimator must has predict')
        return

    # 所有执行fit的操作使用clone一个新的
    estimator = clone(estimator)
    kf = KFold(len(y), n_folds=n_folds, shuffle=True)
    # 首先copy一个一摸一样的y
    y_pred = y.copy()
    """
        eg: y_pred
        array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
               2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
               2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2])
    """

    for train_index, test_index in kf:
        x_train, x_test = x[train_index], x[test_index]
        y_train = y[train_index]
        estimator.fit(x_train, y_train)
        # 通过 estimator.predict(x_test)将y_pred中的值逐步替换
        y_pred[test_index] = estimator.predict(x_test)
    return y_pred



from ultron.utilities.logger import kd_logger
from ultron.factor.data.processing import factor_processing
from ultron.factor.data.winsorize import winsorize_normal
from ultron.tradingday import *
import numpy as np
import pandas as pd
import copy, math
import warnings

warnings.filterwarnings("ignore")


class MultipleFactor(object):

    def __init__(self,
                 signal_data=None,
                 volatility_data=None,
                 returns_data=None):
        self._signal_data = signal_data
        self._volatility_data = volatility_data
        self._returns_data = returns_data

    def _weighted(self, data, equal=0, name='risk'):
        if equal == 0:
            weighted = data[[name]] / data[[name]].sum()
            weighted['code'] = data['code'].values

        else:
            weighted = 1 / len(data)
            weighted = pd.DataFrame([weighted for i in range(0, len(data))],
                                    columns=[name],
                                    index=data.index)
        weighted['code'] = data['code'].values
        weighted = weighted.reset_index().drop(['trade_date'], axis=1)
        return weighted.rename(columns={name: 'weight'})

    def _mixed_signal(self, columns):
        mixed_signal = copy.deepcopy(self._signal_data)
        mixed_signal['signal'] = np.nanmean(mixed_signal.set_index(
            ['trade_date', 'code'])[columns],
                                            axis=1)
        mixed_signal = mixed_signal[['trade_date', 'code', 'signal']]
        ### 过滤叠加信号为Nan
        mixed_signal = mixed_signal.dropna(subset=['signal'])
        return mixed_signal

    def _transformer(self, signal):
        s = copy.deepcopy(signal).sort_values(by=['signal'])
        equal = s[s['signal'] == 0]
        nequal = s[s['signal'] != 0]
        le = len(nequal)
        short = nequal[:int(le / 2)]
        long = nequal[int(le / 2):]
        short['signal'] = short['signal'].apply(lambda x: x
                                                if x < 0 else 0 - x)
        long['signal'] = long['signal'].apply(lambda x: x if x > 0 else 0 - x)
        return pd.concat(
            [long, short, equal],
            axis=0).sort_values(by=['signal']).reset_index(drop=True)

    def _symmetry(self, mixed_signal):
        signal = mixed_signal.set_index('trade_date').groupby(
            level=['trade_date']).apply(lambda x: self._transformer(x))
        return signal.reset_index().drop(['level_1'], axis=1)

    def _initialize_weight(self, mixed_signal, volatility_name):
        data = mixed_signal.merge(self._volatility_data,
                                  on=['trade_date', 'code'])
        data['risk'] = data['signal'] / data[volatility_name]

        ## 权重归一化
        #weighted = data.set_index(['trade_date']).groupby(level=['trade_date']).apply(
        #    lambda x: self._weighted(x)).reset_index()
        #weighted = weighted[['trade_date','weight','code']]
        ### 多空组合，归零处理
        long_weighted = data[data.risk > 0][['trade_date', 'code',
                                             'risk']]  ## 多头权重
        short_weighted = data[data.risk < 0][['trade_date', 'code',
                                              'risk']]  ## 空头权重
        equal_weighted = data[data.risk == 0][['trade_date', 'code',
                                               'risk']]  ### 0 权重
        long_weighted = long_weighted.set_index([
            'trade_date'
        ]).groupby(level=['trade_date']).apply(lambda x: self._weighted(
            data=x, equal=0, name='risk')).reset_index()[[
                'trade_date', 'weight', 'code'
            ]]

        short_weighted = short_weighted.set_index([
            'trade_date'
        ]).groupby(level=['trade_date']).apply(lambda x: self._weighted(
            data=x, equal=0, name='risk')).reset_index()[[
                'trade_date', 'weight', 'code'
            ]]
        short_weighted['weight'] = 0 - short_weighted['weight']

        equal_weighted = equal_weighted[['trade_date', 'risk', 'code'
                                         ]].rename(columns={'risk': 'weight'})
        ### 数据合并
        weighted = long_weighted.append(short_weighted).append(equal_weighted)
        weighted['trade_date'] = pd.to_datetime(weighted['trade_date'])
        return weighted

    def _returns_corr(self, returns):
        return returns.set_index(
            ['trade_date',
             'code'])['nxt1_ret'].unstack().corr(method='spearman')

    def _winsorize_volatility(self, name, volatility_data):
        cols = [name]
        diff_cols = [col for col in volatility_data.columns if col not in cols]
        grouped = volatility_data.groupby(['trade_date'])
        alpha_res = []
        for k, g in grouped:
            new_factors = factor_processing(g[cols].values,
                                            pre_process=[winsorize_normal])
            f = pd.DataFrame(new_factors, columns=cols)
            for k in diff_cols:
                f[k] = g[k].values
            alpha_res.append(f)
        return pd.concat(alpha_res).sort_values(by=['trade_date', 'code'])

    def _weighted_normalised(self, target_pos, windows):
        sum_weighted = target_pos.groupby(
            level=['trade_date']).apply(lambda x: x['weight'].abs().sum())
        ma_weighted = sum_weighted.rolling(windows).mean()
        ma_weighted = (ma_weighted / 2).fillna(0.5)
        ma_weighted.name = 'coeff'
        ### 滚动权重
        ma_weighted = ma_weighted.reset_index()
        v = target_pos.dropna(subset=['weight']).reset_index().merge(
            ma_weighted, on=['trade_date'], how='left')
        v['weight'] = v['weight'] * v['coeff']
        v = v.drop(['coeff'], axis=1).set_index(['trade_date', 'code'])
        return v

    ### 权重计算
    def weighted_run(self,
                     columns=None,
                     default_volatility=0.1,
                     volatility_name='STD20D',
                     period=20,
                     is_symmetry=True,
                     windows=60):
        kd_logger.info("starting construction")
        columns = columns if columns is not None else [col for col in self._signal_data.columns if col \
            not in ['trade_date','code']]
        mixed_signal = self._mixed_signal(columns)
        if is_symmetry:
            mixed_signal = self._symmetry(mixed_signal)

        kd_logger.info("initialize weight")

        weighted = self._initialize_weight(mixed_signal, volatility_name)

        weighted = weighted.sort_values(by=['trade_date', 'code'])
        weighted['trade_date'] = pd.to_datetime(weighted['trade_date'])

        returns_data = self._returns_data.sort_values(
            by=['trade_date', 'code'])
        returns_data['trade_date'] = pd.to_datetime(
            self._returns_data['trade_date'])

        mixed_signal['trade_date'] = pd.to_datetime(mixed_signal['trade_date'])
        mixed_signal = mixed_signal.sort_values(by=['trade_date', 'code'])

        ### 波动率去极值
        kd_logger.info("volatility winsorize")
        volatility_data = self._winsorize_volatility(
            volatility_data=self._volatility_data, name=volatility_name
        )  #self._volatility_data.sort_values(by=['trade_date','code'])
        volatility_data['trade_date'] = pd.to_datetime(
            volatility_data['trade_date'])
        weighted_groups = weighted.groupby('trade_date')
        res = []
        for ref_date, this_data in weighted_groups:
            begin_date = advanceDateByCalendar('china.sse', ref_date,
                                               '-{0}b'.format(period))
            end_date = advanceDateByCalendar('china.sse', ref_date, '-0b')
            kd_logger.info("begin_date: {0}, end_date: {1}".format(
                begin_date, end_date))
            signal = mixed_signal.set_index('trade_date').loc[end_date]
            volatility = volatility_data.set_index('trade_date').loc[end_date]
            returns = returns_data.set_index(
                'trade_date').loc[begin_date:end_date].reset_index()
            codes = set(signal.code.unique().tolist()) & set(
                returns.code.unique().tolist()) & set(
                    volatility.code.unique().tolist())
            returns = returns.set_index('code').loc[codes].reset_index()
            signal = signal.set_index('code').loc[codes].reset_index()
            w = copy.deepcopy(this_data)
            #kd_logger.info("calc returns correlation begin_date:{0}, end_date:{1}".format(
            #    begin_date, end_date))
            corr_dt = self._returns_corr(returns).fillna(0)

            ###重置w, 波动率顺序
            w = w.set_index('code').reindex(corr_dt.index).reset_index()
            volatility = volatility.set_index('code').reindex(
                corr_dt.index).reset_index()
            data = w.merge(corr_dt, on=['code']).merge(volatility, on=['code'])
            cols = [
                col for col in data.columns if col not in
                ['code', 'signal', 'trade_date', 'weight', volatility_name]
            ]
            #kd_logger.info("calc target volatility")
            s = data['weight'] * data[volatility_name]
            v = data[volatility_name]
            n = np.dot(s.T, data[cols])
            if n.shape[0] != s.shape[0]:
                print(n.shape[0], n.shape[0])
            else:
                m = np.dot(n, s)
                if m == 0:
                    continue
                #op = 1 / (s/v).sum() * m
                op = math.sqrt(m)
                weighted_data = copy.deepcopy(this_data)
                weighted_data['weight'] = ((default_volatility / op) *
                                           this_data['weight'])
                res.append(weighted_data.set_index(['trade_date', 'code']))
        target_pos = pd.concat(res, axis=0)
        return target_pos.reset_index(
        ) if not windows > 0 else self._weighted_normalised(
            target_pos, windows).reset_index()

    def equal_run(self, columns=None, equal=1, is_symmetry=True):
        kd_logger.info("starting construction")
        columns = columns if columns is not None else [col for col in self._signal_data.columns if col \
            not in ['trade_date','code']]
        mixed_signal = self._mixed_signal(columns)
        if is_symmetry:
            mixed_signal = self._symmetry(mixed_signal)
        ###权重换算
        long_signal = mixed_signal[mixed_signal.signal > 0]
        short_signal = mixed_signal[mixed_signal.signal < 0]
        equal_signal = mixed_signal[mixed_signal.signal == 0]

        ## weight
        long_weighted = long_signal.set_index([
            'trade_date'
        ]).groupby(level=['trade_date']).apply(lambda x: self._weighted(
            data=x, equal=equal, name='signal')).reset_index()[[
                'trade_date', 'weight', 'code'
            ]]

        short_weighted = short_signal.set_index([
            'trade_date'
        ]).groupby(level=['trade_date']).apply(lambda x: self._weighted(
            data=x, equal=equal, name='signal')).reset_index()[[
                'trade_date', 'weight', 'code'
            ]]

        equal_weighted = equal_signal[['trade_date', 'code']]
        equal_weighted['weight'] = 0.0
        short_weighted['weight'] = 0 - short_weighted['weight']
        weighted = long_weighted.append(short_weighted).append(equal_weighted)
        weighted['trade_date'] = pd.to_datetime(weighted['trade_date'])
        return weighted

    ### 收益计算
    def returns_run(self, target_position, forward_returns):
        data = pd.concat([target_position, forward_returns],
                         axis=1,
                         join='inner')
        target_position = target_position.reindex(data.dropna().index)
        forward_returns = forward_returns.reindex(data.dropna().index)
        weights = target_position.reindex(forward_returns.index)['weight']
        leverags = weights.groupby('trade_date').apply(sum)
        weighted_returns = (np.exp(forward_returns['nxt1_ret']) - 1).multiply(
            weights, axis=0)
        factor_ret_se = weighted_returns.groupby(level='trade_date').sum()
        factor_ret_se = np.log(1 + factor_ret_se)
        return factor_ret_se, leverags, weights



# -*- encoding:utf-8 -*-
"""
    多支交易对象进行择时操作封装模块，内部通过PickTimeWorker进行
    择时，包装完善前后工作，包括多进程下的进度显示，错误处理捕获，结果
    处理等事务
"""
import numpy as np
import pandas as pd
from enum import Enum
from ultron.utilities.logger import kd_logger
from ultron.ump.core.process import add_process_env_sig
from ultron.kdutils.progress import MulPidProgress
from ultron.ump.alpha.pick_time_worker import PickTimeWork
from ultron.ump.alpha import trade_proxy
from ultron.ump.trade import trade_execute


class EFitError(Enum):
    """
        择时操作的错误码
    """

    # 择时操作正常完成，且至少生成一个order
    FIT_OK = 0
    # 择时对象数据获取错误
    NET_ERROR = 1
    # 择时对象数据错误
    DATE_ERROR = 2
    # 择时操作正常完成，但没有生成一个order
    NO_ORDER_GEN = 3
    # 其它错误
    OTHER_ERROR = 4


def _do_pick_time_work(capital,
                       buy_factors,
                       sell_factors,
                       kl_pd,
                       benchmark,
                       show_info=False,
                       show_pg=False):
    """
    内部方法：包装AbuPickTimeWorker进行fit，分配错误码，通过trade_summary生成orders_pd，action_pd
    :param capital: AbuCapital实例对象
    :param buy_factors: 买入因子序列
    :param sell_factors: 卖出因子序列
    :param kl_pd: 金融时间序列
    :param benchmark: 交易基准对象，AbuBenchmark实例对象
    :param draw: 是否绘制在对应的金融时间序列上的交易行为
    :param show_info: 是否显示在整个金融时间序列上的交易结果
    :param show_pg: 是否择时内部启动进度条，适合单进程或者每个进程里只有一个symbol进行择时
    :return:
    """
    if kl_pd is None or kl_pd.shape[0] == 0:
        return None, EFitError.NET_ERROR

    pick_timer_worker = PickTimeWork(capital, kl_pd, benchmark, buy_factors,
                                     sell_factors)
    if show_pg:
        pick_timer_worker.enable_task_pg()
    pick_timer_worker.fit()

    if len(pick_timer_worker.orders) == 0:
        # 择时金融时间序列拟合操作后，没有任何order生成
        return None, EFitError.NO_ORDER_GEN

    # 生成关键的orders_pd与action_pd
    orders_pd, action_pd, _ = trade_proxy.trade_summary(
        pick_timer_worker.orders, kl_pd, show_info=show_info)

    # 最后生成list是因为tuple无法修改导致之后不能灵活处理
    return [orders_pd, action_pd], EFitError.FIT_OK


@add_process_env_sig
def do_symbols_with_same_factors(target_symbols,
                                 benchmark,
                                 buy_factors,
                                 sell_factors,
                                 capital,
                                 apply_capital=True,
                                 kl_pd_manager=None,
                                 show=False,
                                 back_target_symbols=None,
                                 func_factors=None,
                                 show_progress=True):
    """
    输入为多个择时交易对象，以及相同的择时买入，卖出因子序列，对多个交易对象上实施相同的因子
    :param target_symbols: 多个择时交易对象序列
    :param benchmark: 交易基准对象，AbuBenchmark实例对象
    :param buy_factors: 买入因子序列
    :param sell_factors: 卖出因子序列
    :param capital: AbuCapital实例对象
    :param apply_capital: 是否进行资金对象的融合，多进程环境下将是False
    :param kl_pd_manager: 金融时间序列管理对象，AbuKLManager实例
    :param show: 是否显示每个交易对象的交易细节
    :param back_target_symbols:  补位targetSymbols为了忽略网络问题及数据不足导致的问题
    :param func_factors: funcFactors在内层解开factors dicts为了do_symbols_with_diff_factors
    :param show_progress: 进度条显示，默认True
    """

    def _batch_symbols_with_same_factors(p_buy_factors, p_sell_factors):
        r_orders_pd = None
        r_action_pd = None
        r_all_fit_symbols_cnt = 0
        # 启动多进程进度显示AbuMulPidProgress
        with MulPidProgress(len(target_symbols),
                            'pick times complete',
                            show_progress=show_progress) as progress:
            for epoch, target_symbol in enumerate(target_symbols):
                # 如果symbol只有一个就不show了，留给下面_do_pick_time_work中show_pg内部显示进度
                if len(target_symbols) > 1:
                    # 如果要绘制交易细节就不要clear了
                    progress.show(epoch + 1, clear=not show)

                if func_factors is not None and callable(func_factors):
                    # 针对do_symbols_with_diff_factors mul factors等情况嵌入可变因子
                    p_buy_factors, p_sell_factors = func_factors(target_symbol)
                try:
                    kl_pd = kl_pd_manager.get_pick_time_kl_pd(target_symbol)
                    ret, fit_error = _do_pick_time_work(
                        capital,
                        p_buy_factors,
                        p_sell_factors,
                        kl_pd,
                        benchmark,
                        show_info=show,
                        show_pg=(len(target_symbols) == 1 and show_progress))
                except Exception as e:
                    kd_logger.exception("{0}:{1}".format(target_symbol, e))
                    continue

                if ret is None and back_target_symbols is not None:
                    # 择时结果错误或者没有order生成的情况下，如果有补位序列，择从序列中pop出一个，进行补位
                    if fit_error is not None and fit_error == EFitError.NO_ORDER_GEN:
                        # 没有order生成的要统计进去
                        r_all_fit_symbols_cnt += 1
                    while True:
                        if len(back_target_symbols) <= 0:
                            break
                        # pop出来代替原先的target
                        target_symbol = back_target_symbols.pop()
                        kl_pd = kl_pd_manager.get_pick_time_kl_pd(
                            target_symbol)
                        ret, fit_error = _do_pick_time_work(capital,
                                                            p_buy_factors,
                                                            p_sell_factors,
                                                            kl_pd,
                                                            benchmark,
                                                            show_info=show)
                        if fit_error == EFitError.NO_ORDER_GEN:
                            r_all_fit_symbols_cnt += 1
                        if ret is not None:
                            break
                if ret is None:
                    continue
                r_all_fit_symbols_cnt += 1
                # 连接每一个交易对象生成的orders_pd和action_pd
                r_orders_pd = ret[0] if r_orders_pd is None else pd.concat(
                    [r_orders_pd, ret[0]])
                r_action_pd = ret[1] if r_action_pd is None else pd.concat(
                    [r_action_pd, ret[1]])
        return r_orders_pd, r_action_pd, r_all_fit_symbols_cnt

    orders_pd, action_pd, all_fit_symbols_cnt = _batch_symbols_with_same_factors(
        buy_factors, sell_factors)
    if orders_pd is not None and action_pd is not None:

        # 要sort'Date', 'action'两项，不然之后的行apply_action_to_capital后有问题
        # noinspection PyUnresolvedReferences
        action_pd = action_pd.sort_values(['Date', 'action'])
        action_pd.index = np.arange(0, action_pd.shape[0])
        # noinspection PyUnresolvedReferences
        orders_pd = orders_pd.sort_values(['buy_date'])
        if apply_capital:
            # 如果非多进程环境下开始融合资金对象
            trade_execute.apply_action_to_capital(capital,
                                                  action_pd,
                                                  kl_pd_manager,
                                                  show_progress=show_progress)

    return orders_pd, action_pd, all_fit_symbols_cnt



# -*- coding: utf-8 -*-
import unittest

import numpy as np

from ultron.optimize.linearbuilder import linear_builder

class LinearBuild(unittest.TestCase):
    def setUp(self):
        self.er = np.random.randn(3000)
        self.risk_exp = np.random.randn(3000, 30)
        self.risk_exp = np.concatenate([self.risk_exp, np.ones((3000, 1))], axis=1)
        self.bm = np.random.randint(100, size=3000).astype(float)
        self.current_pos = np.random.randint(0, 100, size=3000)
        self.current_pos = self.current_pos / self.current_pos.sum()

    def test_linear_build(self):
        bm = self.bm / self.bm.sum()
        eplson = 1e-6

        status, _, w = linear_builder(self.er,
                                0.,
                                0.01,
                                self.risk_exp,
                                (bm @ self.risk_exp, bm @ self.risk_exp))
        self.assertEqual(status, 'optimal')
        self.assertAlmostEqual(np.sum(w), 1.)
        self.assertTrue(np.all(w <= 0.01 + eplson))
        self.assertTrue(np.all(w >= -eplson))

        calc_risk = (w - bm) @ self.risk_exp
        expected_risk = np.zeros(self.risk_exp.shape[1])
        np.testing.assert_array_almost_equal(calc_risk, expected_risk)

    def test_linear_build_with_inequality_constraints(self):
        bm = self.bm / self.bm.sum()
        eplson = 1e-6

        turn_over_target = 0.1

        risk_lbound = bm @ self.risk_exp
        risk_ubound = bm @ self.risk_exp

        risk_tolerance = 0.01 * np.abs(risk_lbound[:-1])

        risk_lbound[:-1] = risk_lbound[:-1] - risk_tolerance
        risk_ubound[:-1] = risk_ubound[:-1] + risk_tolerance

        status, _, w = linear_builder(self.er,
                                      0.,
                                      0.01,
                                      self.risk_exp,
                                      risk_target=(risk_lbound, risk_ubound),
                                      turn_over_target=turn_over_target,
                                      current_position=self.current_pos)

        self.assertEqual(status, 'optimal')
        self.assertAlmostEqual(np.sum(w), 1.)
        self.assertTrue(np.all(w <= 0.01 + eplson))
        self.assertTrue(np.all(w >= -eplson))

        self.assertAlmostEqual(np.abs(w - self.current_pos).sum(), turn_over_target)

        calc_risk = (w - bm) @ self.risk_exp / np.abs(bm @ self.risk_exp)
        self.assertTrue(np.all(np.abs(calc_risk) <= 1.0001e-2))
    
    def test_linear_build_with_to_constraint_with_ecos(self):
        bm = self.bm / self.bm.sum()
        eplson = 1e-6
        turn_over_target = 0.1

        risk_lbound = bm @ self.risk_exp
        risk_ubound = bm @ self.risk_exp

        risk_tolerance = 0.01 * np.abs(risk_lbound[:-1])

        risk_lbound[:-1] = risk_lbound[:-1] - risk_tolerance
        risk_ubound[:-1] = risk_ubound[:-1] + risk_tolerance

        status, _, w = linear_builder(self.er,
                                      0.,
                                      0.01,
                                      self.risk_exp,
                                      risk_target=(risk_lbound, risk_ubound),
                                      turn_over_target=turn_over_target,
                                      current_position=self.current_pos,
                                      method='ecos')
        self.assertEqual(status, 'optimal')
        self.assertAlmostEqual(np.sum(w), 1.)
        self.assertTrue(np.all(w <= 0.01 + eplson))
        self.assertTrue(np.all(w >= -eplson))
        self.assertAlmostEqual(np.abs(w - self.current_pos).sum(), turn_over_target)

        calc_risk = (w - bm) @ self.risk_exp / np.abs(bm @ self.risk_exp)
        self.assertTrue(np.all(np.abs(calc_risk) <= 1.0001e-2))


# -*- encoding:utf-8 -*-
"""
    横截面下因子评估方式--因子状态
"""
import pdb, functools
import numpy as np
import pandas as pd
from collections import namedtuple
from ultron.utilities.logger import kd_logger
from ultron.kdutils.progress import Progress


class StateTuple(
        namedtuple('StateTuple',
                   ('mean', 'median', 'std', 'max', 'min', 'nanr', 'zeror'))):
    __slots__ = ()

    def __repr__(self):
        return "\nmean:{}\nmedian:{}\nstd:{}\nmax:{}\nmin:{}\nnanr:{}\nzeror:{}".format(
            self.mean, self.median, self.std, self.max, self.min, self.nanr,
            self.zeror)


def valid_check(func):
    """检测度量的输入是否正常，非正常显示info，正常继续执行被装饰方法"""

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.valid:
            return func(self, *args, **kwargs)
        else:
            kd_logger.info('metrics input is invalid or zero order gen!')

    return wrapper


class State(object):

    @classmethod
    def general(cls, factors, dummy, show_log=True):
        return cls(factors, dummy, show_log).fit_state()

    def __init__(self, factors, dummy=None, show_log=True):
        self.valid = False
        self.factors = factors
        self.dummy = dummy
        self.show_log = show_log
        if self.factors is not None and self.dummy is not None:
            self.valid = True

    @valid_check
    def fit_state(self):
        total_count = np.nansum(self.dummy.values)
        factors_dummy = self.factors.values * self.dummy.values

        nan_mean = np.nanmean(factors_dummy)
        nan_median = np.nanmedian(factors_dummy)
        nan_std = np.nanstd(factors_dummy)
        nan_max = np.nanmax(factors_dummy)
        nan_min = np.nanmin(factors_dummy)

        nanr = 1 - np.sum(abs(factors_dummy) >= 0) / total_count
        zeror = np.sum(factors_dummy == 0) / total_count
        return StateTuple(mean=nan_mean,
                          median=nan_median,
                          std=nan_std,
                          max=nan_max,
                          min=nan_min,
                          nanr=nanr,
                          zeror=zeror)



# -*- encoding:utf-8 -*-
# 基础库导入

import os, sys
import numpy as np
import pandas as pd

sys.path.insert(0, os.path.abspath('../'))

from ultron.env import *
from ultron.ump.factor import FactorPreAtrNStop, FactorAtrNStop, FactorCloseAtrNStop, FactorBuyBreak

from ultron.ump.metrics import grid_helper as GridHelper
from ultron.ump.metrics.grid_search import GridSearch

enable_example_env()

stop_win_range = np.arange(2.0, 2.5, 0.5)
stop_loss_range = np.arange(0.5, 1, 0.5)

sell_atr_nstop_factor_grid = {
    'class': [FactorAtrNStop],
    'stop_loss_n': stop_loss_range,
    'stop_win_n': stop_win_range
}

print('FactorAtrNStop止盈参数stop_win_n设置范围:{}'.format(stop_win_range))
print('FactorAtrNStop止损参数stop_loss_n设置范围:{}'.format(stop_loss_range))

close_atr_range = np.arange(1.0, 1.5, 0.5)
pre_atr_range = np.arange(1.0, 1.5, 0.5)

sell_atr_pre_factor_grid = {
    'class': [FactorPreAtrNStop],
    'pre_atr_n': pre_atr_range
}

sell_atr_close_factor_grid = {
    'class': [FactorCloseAtrNStop],
    'close_atr_n': close_atr_range
}

print('暴跌保护止损参数pre_atr_n设置范围:{}'.format(pre_atr_range))
print('盈利保护止盈参数close_atr_n设置范围:{}'.format(close_atr_range))

sell_factors_product = GridHelper.gen_factor_grid(
    GridHelper.K_GEN_FACTOR_PARAMS_SELL, [
        sell_atr_nstop_factor_grid, sell_atr_pre_factor_grid,
        sell_atr_close_factor_grid
    ],
    need_empty_sell=True)

print('卖出因子参数共有{}种组合方式'.format(len(sell_factors_product)))
print('卖出因子组合0: 形式为{}'.format(sell_factors_product[0]))

sell_factors_product = GridHelper.gen_factor_grid(
    GridHelper.K_GEN_FACTOR_PARAMS_SELL, [
        sell_atr_nstop_factor_grid, sell_atr_pre_factor_grid,
        sell_atr_close_factor_grid
    ],
    need_empty_sell=True)

print('卖出因子参数共有{}种组合方式'.format(len(sell_factors_product)))
print('卖出因子组合0: 形式为{}'.format(sell_factors_product[0]))

buy_bk_factor_grid1 = {'class': [FactorBuyBreak], 'xd': [42]}

buy_bk_factor_grid2 = {'class': [FactorBuyBreak], 'xd': [60]}

buy_factors_product = GridHelper.gen_factor_grid(
    GridHelper.K_GEN_FACTOR_PARAMS_BUY,
    [buy_bk_factor_grid1, buy_bk_factor_grid2])

print('买入因子参数共有{}种组合方式'.format(len(buy_factors_product)))
print('买入因子组合形式为{}'.format(buy_factors_product))

read_cash = 1000000
choice_symbols = ['usAAPL', 'usBIDU', 'usFB', 'usTSLA']

#### 加载指数行情
pick_kl_pd_dict = {}

benchmark_kl_pd = pd.read_csv(os.path.join('data',
                                           'us.IXIC_20120723_20160726.csv'),
                              index_col=0)
benchmark_kl_pd.index = pd.to_datetime(benchmark_kl_pd.index)
benchmark_kl_pd.name = 'us_NYSE:.IXIC'

aaple_kl_pd = pd.read_csv(os.path.join('data', 'usAAPL_20140609_20160726.csv'),
                          index_col=0)
aaple_kl_pd.index = pd.to_datetime(aaple_kl_pd.index)
aaple_kl_pd.name = 'usAAPL'
pick_kl_pd_dict['usAAPL'] = aaple_kl_pd

bidu_kl_pd = pd.read_csv(os.path.join('data', 'usBIDU_20130726_20160726.csv'),
                         index_col=0)
bidu_kl_pd.index = pd.to_datetime(bidu_kl_pd.index)
bidu_kl_pd.name = 'usBIDU'
pick_kl_pd_dict['usBIDU'] = bidu_kl_pd

fb_kl_pd = pd.read_csv(os.path.join('data', 'usFB_20140723_20160726.csv'),
                       index_col=0)
fb_kl_pd.index = pd.to_datetime(fb_kl_pd.index)
fb_kl_pd.name = 'usFB'
pick_kl_pd_dict['usFB'] = fb_kl_pd

tsla_kl_pd = pd.read_csv(os.path.join('data', 'usTSLA_20130726_20160726.csv'),
                         index_col=0)
tsla_kl_pd.index = pd.to_datetime(tsla_kl_pd.index)
tsla_kl_pd.name = 'usTSLA'
pick_kl_pd_dict['usTSLA'] = tsla_kl_pd

grid_search = GridSearch(read_cash,
                         choice_symbols,
                         benchmark_kl_pd=benchmark_kl_pd,
                         buy_factors_product=buy_factors_product,
                         sell_factors_product=sell_factors_product)

#### 外部设置K线
grid_search.kl_pd_manager.set_pick_time(pick_kl_pd_dict)

grid_search.fit(n_jobs=1)


# -*- encoding:utf-8 -*-
"""
    买入择时示例因子：长线趋势跟踪与短线均值回复的配合
"""

from ultron.ump.factor.buy.base import BuyCallMixin, FactorBuyXD
from ultron.ump.technical.line import Line


class UpDownTrend(FactorBuyXD, BuyCallMixin):
    """示例长线上涨中寻找短线下跌买入择时因子，混入BuyCallMixin"""

    def _init_self(self, **kwargs):
        """
            kwargs中可以包含xd: 比如20，30，40天...突破，默认20
            kwargs中可以包含past_factor: 代表长线的趋势判断长度，默认4，long = xd * past_factor->eg: long = 20 * 4
            kwargs中可以包含up_deg_threshold: 代表判断上涨趋势拟合角度阀值，即长线拟合角度值多少决策为上涨，默认3
        """
        if 'xd' not in kwargs:
            # 如果外部没有设置xd值，默认给一个30
            kwargs['xd'] = 20
        super(UpDownTrend, self)._init_self(**kwargs)
        # 代表长线的趋势判断长度，默认4，long = xd * past_factor->eg: long = 30 * 4
        self.past_factor = kwargs.pop('past_factor', 4)
        # 代表判断上涨趋势拟合角度阀值，即长线拟合角度值多少决策为上涨，默认4
        self.up_deg_threshold = kwargs.pop('up_deg_threshold', 3)

    def fit_day(self, today):
        """
        长线周期选择目标为上升趋势的目标，短线寻找近期走势为向下趋势的目标进行买入，期望是持续之前长相的趋势
            1. 通过past_today_kl获取长周期的金融时间序列，通过AbuTLine中的is_up_trend判断
            长周期是否属于上涨趋势，
            2. 今天收盘价为最近xd天内最低价格，且短线xd天的价格走势为下跌趋势
            3. 满足1，2发出买入信号
        :param today: 当前驱动的交易日金融时间序列数据
        """
        long_kl = self.past_today_kl(today, self.past_factor * self.xd)
        tl_long = Line(long_kl.close, 'long')
        # 判断长周期是否属于上涨趋势
        if tl_long.is_up_trend(up_deg_threshold=self.up_deg_threshold):
            if today.close == self.xd_kl.close.min() and Line(
                    self.xd_kl.close, 'short').is_down_trend(
                        down_deg_threshold=-self.up_deg_threshold):
                # 今天收盘价为最近xd天内最低价格，且短线xd天的价格走势为下跌趋势
                return self.buy_tomorrow()


class DownUpTrend(FactorBuyXD, BuyCallMixin):
    """示例长线下跌中寻找短线突破反转买入择时因子，混入BuyCallMixin，即向上突破触发买入event"""

    def _init_self(self, **kwargs):
        """
            kwargs中可以包含xd: 比如20，30，40天...突破，默认20
            kwargs中可以包含past_factor: 代表长线的趋势判断长度，默认4，long = xd * past_factor->eg: long = 20 * 4
            kwargs中可以包含down_deg_threshold: 代表判断下跌趋势拟合角度阀值，即长线拟合角度值多少决策为下跌，默认-3
        """
        if 'xd' not in kwargs:
            # 如果外部没有设置xd值，默认给一个20
            kwargs['xd'] = 20

        super(DownUpTrend, self)._init_self(**kwargs)
        # 代表长线的趋势判断长度，默认4，long = xd * past_factor->eg: long = 20 * 4
        self.past_factor = kwargs.pop('past_factor', 4)
        # 代表判断下跌趋势拟合角度阀值，即长线拟合角度值多少决策为下跌，默认-3
        self.down_deg_threshold = kwargs.pop('down_deg_threshold', -3)

    def fit_day(self, today):
        """
        长线下跌中寻找短线突破反转买入择时因子
            1. 通过past_today_kl获取长周期的金融时间序列，通过AbuTLine中的is_down_trend判断
            长周期是否属于下跌趋势，
            2. 今天收盘价为最近xd天内最高价格，且短线xd天的价格走势为上升趋势
            3. 满足1，2发出买入信号
        :param today: 当前驱动的交易日金融时间序列数据
        """
        long_kl = self.past_today_kl(today, self.past_factor * self.xd)
        tl_long = Line(long_kl.close, 'long')
        # 判断长周期是否属于下跌趋势
        if tl_long.is_down_trend(down_deg_threshold=self.down_deg_threshold):
            if today.close == self.xd_kl.close.max() and Line(
                    self.xd_kl.close, 'short').is_up_trend(
                        up_deg_threshold=-self.down_deg_threshold):
                # 今天收盘价为最近xd天内最高价格，且短线xd天的价格走势为上升趋势
                return self.buy_tomorrow()


# -*- coding: utf-8 -*-
import time, datetime, pdb
import numpy as np
from ultron.optimize.geneticist.genetic import Gentic

MAX_INT = np.iinfo(np.int32).max
MIN_INT = np.iinfo(np.int32).min


class Engine(object):

    def __init__(
            self,
            model_sets,
            params_sets,
            population_size=2000,
            generations=MAX_INT,
            tournament_size=20,
            n_jobs=1,
            p_crossover=0.9,
            p_point_mutation=0.01,
            p_subtree_mutation=0.00,
            p_hoist_mutation=0.00,
            p_point_replace=0.8,
            stopping_criteria=0.0,
            fitness=None,
            save_model=None,
            random_state=None,
            greater_is_better=True,  #True 倒序， False 正序
            convergence=None,
            standard_score=0.9,  # None代表 根据tournament_size保留种群  standard_score保留种群
            custom_params=None,
            rootid=0,
            verbose=1):

        self._population_size = population_size
        self._generations = MAX_INT if generations == 0 else generations
        self._tournament_size = tournament_size
        self._stopping_criteria = stopping_criteria
        self.init_method = 'full'
        self._model_sets = model_sets
        self._params_sets = params_sets
        self._p_crossover = p_crossover
        self._p_subtree_mutation = p_subtree_mutation
        self._p_hoist_mutation = p_hoist_mutation
        self._p_point_mutation = p_point_mutation
        self._p_point_replace = p_point_replace
        self._random_state = random_state
        self._fitness = fitness
        self._save_model = save_model
        self._greater_is_better = greater_is_better
        self._standard_score = standard_score
        self._convergence = convergence
        self._custom_params = custom_params
        self._con_time = 0
        self._n_jobs = n_jobs
        self._verbose = verbose
        self._rootid = int(
            time.time() * 1000000 +
            datetime.datetime.now().microsecond) if rootid == 0 else rootid

    def run_gentic(self, features, X, Y, mode, n_splits):
        gentic = Gentic(
            model_sets=self._model_sets,
            params_sets=self._params_sets,
            population_size=self._population_size,
            generations=self._generations,
            tournament_size=self._tournament_size,
            n_jobs=self._n_jobs,
            p_crossover=self._p_crossover,
            p_point_mutation=self._p_point_mutation,
            p_point_replace=self._p_point_replace,
            stopping_criteria=self._stopping_criteria,
            fitness=self._fitness,
            save_model=self._save_model,
            custom_params=self._custom_params,
            random_state=self._random_state,
            greater_is_better=self._greater_is_better,  #True 倒序， False 正序
            convergence=self._convergence,
            rootid=self._rootid,
            standard_score=self.
            _standard_score,  # None代表 根据tournament_size保留种群  standard_score保留种群
            verbose=self._verbose)
        gentic.train(features=features, X=X, Y=Y, mode=mode, n_splits=n_splits)
        result = gentic._run_details
        raw_fitness = 0 if len(
            result['best_programs']) == 0 else result['best_fitness'][-1]
        # del gentic 主动释放内存
        return raw_fitness

    def train(self, features, X, Y, mode, n_splits):
        raw_fitness = 0
        while raw_fitness <= self._stopping_criteria if self._greater_is_better else raw_fitness >= self._stopping_criteria:
            raw_fitness = self.run_gentic(features,
                                          X=X,
                                          Y=Y,
                                          mode=mode,
                                          n_splits=n_splits)



# -*- encoding:utf-8 -*-
"""
    示例买入择时因子
"""
from ultron.ump.factor.buy.base import FactorBuyBase, FactorBuyXD, FactorBuyTD, BuyCallMixin
from ultron.ump.factor.buy.buy_break import FactorBuyBreak
from ultron.ump.factor.buy.buy_wrap import LeastPolyWrap
from ultron.ump.technical.line import Line


class FactorSDBreak(FactorBuyXD, BuyCallMixin):
    """示例买入因子： 在FactorBuyBreak基础上进行降低交易频率，提高系统的稳定性处理"""

    def _init_self(self, **kwargs):
        """
        :param kwargs: kwargs可选参数poly值，poly在fit_month中和每一个月大盘计算的poly比较，
        若是大盘的poly大于poly认为走势震荡，poly默认为2
        """
        super(FactorSDBreak, self)._init_self(**kwargs)
        # poly阀值，self.poly在fit_month中和每一个月大盘计算的poly比较，若是大盘的poly大于poly认为走势震荡
        self.poly = kwargs.pop('poly', 2)
        # 是否封锁买入策略进行择时交易
        self.lock = False

    def fit_month(self, today):
        # fit_month即在回测策略中每一个月执行一次的方法
        # 策略中拥有self.benchmark，即交易基准对象，Benchmark实例对象，benchmark.kl_pd即对应的市场大盘走势
        benchmark_df = self.benchmark.kl_pd
        # 拿出大盘的今天
        benchmark_today = benchmark_df[benchmark_df.date == today.date]
        if benchmark_today.empty:
            return 0
        # 要拿大盘最近一个月的走势，准备切片的start，end
        end_key = int(benchmark_today.iloc[0].key)
        start_key = end_key - 20
        if start_key < 0:
            return False
        # 使用切片切出从今天开始向前20天的数据
        benchmark_month = benchmark_df.set_index('key').loc[start_key:end_key +
                                                            1].reset_index()

        # 通过大盘最近一个月的收盘价格做为参数构造TLine对象
        benchmark_month_line = Line(benchmark_month.close,
                                    'benchmark month line')
        # 计算这个月最少需要几次拟合才能代表走势曲线

        least, _, _, _, _, _ = benchmark_month_line.create_least_valid_poly()

        if least >= self.poly:
            # 如果最少的拟合次数大于阀值self.poly，说明走势成立，大盘非震荡走势，解锁交易
            self.lock = False
        else:
            # 如果最少的拟合次数小于阀值self.poly，说明大盘处于震荡走势，封锁策略进行交易
            self.lock = True

    def fit_day(self, today):
        if self.lock:
            # 如果封锁策略进行交易的情况下，策略不进行择时
            return None

        # 今天的收盘价格达到xd天内最高价格则符合买入条件
        if today.close == self.xd_kl.close.max():
            return self.buy_tomorrow()


@LeastPolyWrap()
class TwoDayBuy(FactorBuyTD, BuyCallMixin):
    """示例LeastPolyWrap，混入BuyCallMixin，即向上突破触发买入event"""

    def _init_self(self, **kwargs):
        """简单示例什么都不编写了"""
        pass

    def fit_day(self, today):
        """
        针对每一个交易日拟合买入交易策略，今天涨，昨天涨就买
        :param today: 当前驱动的交易日金融时间序列数据
        :return:
        """
        # 今天的涨幅
        td_change = today.p_change
        # 昨天的涨幅
        yd_change = self.yesterday.p_change

        if td_change > 0 and 0 < yd_change < td_change:
            # 连续涨两天, 且今天的涨幅比昨天还高 －>买入, 用到了今天的涨幅，只能明天买
            return self.buy_tomorrow()
        return None


class WeekMonthBuy(FactorBuyBase, BuyCallMixin):
    """策略示例每周买入一次或者每一个月买入一次"""

    def _init_self(self, **kwargs):
        """kwargs可选参数：is_buy_month，bool默认True一个月买入一次, False一周买入一次"""
        self.is_buy_month = kwargs.pop('is_buy_month', True)

    def fit_day(self, today):
        """
        :param today: 当前驱动的交易日金融时间序列数据
        """
        if self.is_buy_month and today.exec_month or not self.is_buy_month and today.exec_week:
            # 没有用到今天的任何数据，直接今天买入
            return self.buy_today()



# -*- coding: utf-8 -*-

import unittest

import numpy as np

from ultron.factor.data.quantile import quantile


class Quantile(unittest.TestCase):

    def test_quantile(self):
        n = 5000
        bins = 10
        s = np.random.randn(n)
        calculated = quantile(s, bins)

        rank = s.argsort().argsort()

        bin_size = float(n) / bins
        pillars = [int(i * bin_size) for i in range(1, bins + 1)]

        starter = 0
        for i, r in enumerate(pillars):
            self.assertTrue(np.all(calculated[(rank >= starter) & (rank < r)] == i))
            starter = r


# -*- encoding:utf-8 -*-
"""
ATR

ATR又称 Average true range平均真实波动范围，简称ATR指标，是由J.Welles Wilder 发明的，ATR指标主要是用来衡量市场波动的强烈度，
即为了显示市场变化率的指标。

计算方法：
1. TR=∣最高价-最低价∣，∣最高价-昨收∣，∣昨收-最低价∣中的最大值
2. 真实波幅（ATR）= MA(TR,N)（TR的N日简单移动平均）
3. 常用参数N设置为14日或者21日
"""

import pandas as pd
import numpy as np
from ultron.ump.core.helper import pd_ewm_mean
from ultron.kdutils import scaler as scaler_util
from ultron.ump.indicator.base import g_calc_type, ECalcType


def _calc_atr_from_ta(high, low, close, time_period=14):
    """
    使用talib计算atr，即透传talib.ATR计算结果
    :param high: 最高价格序列，pd.Series或者np.array
    :param low: 最低价格序列，pd.Series或者np.array
    :param close: 收盘价格序列，pd.Series或者np.array
    :param time_period: atr的N值默认值14，int
    :return: atr值序列，np.array对象
    """
    import talib
    if isinstance(high, pd.Series):
        high = high.values
    if isinstance(low, pd.Series):
        low = low.values
    if isinstance(close, pd.Series):
        close = close.values

    atr = talib.ATR(high, low, close, timeperiod=time_period)
    return atr


def _calc_atr_from_pd(high, low, close, time_period=14):
    """
    通过atr公式手动计算atr
    :param high: 最高价格序列，pd.Series或者np.array
    :param low: 最低价格序列，pd.Series或者np.array
    :param close: 收盘价格序列，pd.Series或者np.array
    :param time_period: atr的N值默认值14，int
    :return: atr值序列，np.array对象
    """
    if isinstance(close, pd.Series):
        # shift(1)构成昨天收盘价格序列
        pre_close = close.shift(1).values
    else:
        from scipy.ndimage.interpolation import shift
        # 也可以暂时转换为pd.Series进行shift
        pre_close = shift(close, 1)
    pre_close[0] = pre_close[1]

    if isinstance(high, pd.Series):
        high = high.values
    if isinstance(low, pd.Series):
        low = low.values

    # ∣最高价 - 最低价∣
    tr_hl = np.abs(high - low)
    # ∣最高价 - 昨收∣
    tr_hc = np.abs(high - pre_close)
    # ∣昨收 - 最低价∣
    tr_cl = np.abs(pre_close - low)
    # TR =∣最高价 - 最低价∣，∣最高价 - 昨收∣，∣昨收 - 最低价∣中的最大值
    tr = np.maximum(np.maximum(tr_hl, tr_hc), tr_cl)
    # （ATR）= MA(TR, N)（TR的N日简单移动平均）, 这里没有完全按照标准公式使用简单移动平均，使用了pd_ewm_mean，即加权移动平均
    atr = pd_ewm_mean(pd.Series(tr), span=time_period, min_periods=1)
    # 返回atr值序列，np.array对象
    return atr.values


"""通过在NDBase中尝试import talib来统一确定指标计算方式"""
calc_atr = _calc_atr_from_pd if g_calc_type == ECalcType.E_FROM_PD else _calc_atr_from_ta


def atr14(high, low, close):
    """
    通过high, low, close计算atr14序列值
    :param high: 最高价格序列，pd.Series或者np.array
    :param low: 最低价格序列，pd.Series或者np.array
    :param close: 收盘价格序列，pd.Series或者np.array
    :return: atr值序列，np.array对象
    """
    atr = calc_atr(high, low, close, 14)
    return atr


def atr21(high, low, close):
    """
    通过high, low, close计算atr21序列值
    :param high: 最高价格序列，pd.Series或者np.array
    :param low: 最低价格序列，pd.Series或者np.array
    :param close: 收盘价格序列，pd.Series或者np.array
    :return: atr值序列，np.array对象
    """
    atr = calc_atr(high, low, close, 21)
    return atr


def atr14_min(high, low, close):
    """
    确定常数阀值时使用，通过high, low, close计算atr14序列值，返回计算结果atr14序列中的最小值
    :param high: 最高价格序列，pd.Series或者np.array
    :param low: 最低价格序列，pd.Series或者np.array
    :param close: 收盘价格序列，pd.Series或者np.array
    :return: atr值序列，atr14序列中的最小值，float
    """
    _atr14 = atr14(high, low, close)
    _atr14 = pd.Series(_atr14)
    _atr14.fillna(method='bfill', inplace=True)
    _atr14 = _atr14.min()
    return _atr14


def atr14_max(high, low, close):
    """
    确定常数阀值时使用，通过high, low, close计算atr14序列值，返回计算结果atr14序列中的最大值
    :param high: 最高价格序列，pd.Series或者np.array
    :param low: 最低价格序列，pd.Series或者np.array
    :param close: 收盘价格序列，pd.Series或者np.array
    :return: atr值序列，atr14序列中的最大值，float
    """
    _atr14 = atr14(high, low, close)
    _atr14 = pd.Series(_atr14)
    _atr14.fillna(method='bfill', inplace=True)
    _atr14 = _atr14.max()
    return _atr14


def atr21_min(high, low, close):
    """
    确定常数阀值时使用，通过high, low, close计算atr21序列值，返回计算结果atr21序列中的最小值
    :param high: 最高价格序列，pd.Series或者np.array
    :param low: 最低价格序列，pd.Series或者np.array
    :param close: 收盘价格序列，pd.Series或者np.array
    :return: atr值序列，atr21序列中的最小值，float
    """
    _atr21 = atr21(high, low, close)
    _atr21 = pd.Series(_atr21)
    _atr21.fillna(method='bfill', inplace=True)
    _atr21 = _atr21.min()
    return _atr21


def atr21_max(high, low, close):
    """
    确定常数阀值时使用，通过high, low, close计算atr21序列值，返回计算结果atr21序列中的最大值
    :param high: 最高价格序列，pd.Series或者np.array
    :param low: 最低价格序列，pd.Series或者np.array
    :param close: 收盘价格序列，pd.Series或者np.array
    :return: atr值序列，atr21序列中的最大值，float
    """
    _atr21 = atr21(high, low, close)
    _atr21 = pd.Series(_atr21)
    _atr21.fillna(method='bfill', inplace=True)
    _atr21 = _atr21.max()
    return _atr21


# -*- coding: utf-8 -*-
import abc
import pandas as pd
import numpy as np
from ultron.strategy.builder.factor import ts_std_mean0, ts_grade, cross_section_demean_rank, cross_section_scale


def calc_portfolio_net_return(portfolio, return_, commission_rate=None):
    x = portfolio.shift(periods=1, freq=None) * return_
    turnover = portfolio.diff(periods=1) - x
    commission = turnover.abs() * commission_rate
    return x - commission


class LeverageRisk(metaclass=abc.ABCMeta):

    def __init__(self, **kwargs):
        self._risk_list = kwargs['risk_list']
        self._weight_list = kwargs['weight_list']
        self._risk_builder = kwargs['risk_builder']
        self._risk_control = kwargs['risk_control']

    def calc_ratio(self, factors, **kwargs):
        if self._risk_builder.lower() == 'riskweighted':
            rval = self.weighted_ratio(factors)
        else:
            raise NotImplementedError(
                'unknown leverage-risk-builder method {}'.format(
                    self._risk_builder))

        return rval

    def weighted_ratio(self, factors):
        risk_weight_list = self._weight_list
        risk_list = self._risk_list
        if isinstance(risk_weight_list, list):
            risk_weight_list = np.array(risk_weight_list)
            risk_weight_list = pd.DataFrame(risk_weight_list).T
        weights = cross_section_scale(risk_weight_list)
        weights = weights.values[0].tolist()

        index = factors[risk_list[0]].index
        columns = factors[risk_list[0]].columns
        rval = pd.DataFrame(0, index=index, columns=columns)
        for i, risk in enumerate(risk_list):
            rval += weights[i] * factors[risk]
        return rval

    def constrained_leverage_portfolio(self, portfolio):
        rval = portfolio.copy()
        ratio = rval.abs().sum(axis=1)
        ratio[ratio < 1] = 1
        rval = rval.div(ratio + 1e-16, axis=0)
        return rval

    def run(self, portfolio, factors=None):
        if self._risk_builder.lower() == 'riskweighted':
            risk_list = self._risk_list
            risk_weight_list = self._weight_list

            if len(risk_list) > 0:
                ratio = self.weighted_ratio(factors=factors)
                rval = portfolio * ratio
            else:
                rval = portfolio.copy()
        else:
            raise NotImplementedError(
                'unknown leverageRiskBuilder method {0:s}'.format(
                    self._risk_builder))

        if self._risk_control:
            rval = self.constrained_leverage_portfolio(rval)

        return rval


class ExposureRisk(metaclass=abc.ABCMeta):

    def __init__(self, **kwargs):
        self._commission_rate = kwargs['commission_rate']
        self._portfolio_exposure = kwargs['portfolio_exposure']
        self._max_portfolio = kwargs['max_portfolio']
        self._product_exposure = kwargs['product_exposure']
        self._max_product = kwargs['max_product']
        self._turnover_winnratio_winsize = kwargs['turnover_winnratio_winsize']

    def portfolio_constraint_positions(self, portfolio, exposure=None):
        if exposure is None:
            exposure = self._max_portfolio
        rval = portfolio.copy()
        index = portfolio[portfolio.sum(axis=1).abs() > exposure].index
        rval = portfolio.copy()
        index = portfolio[portfolio.sum(axis=1).abs() > exposure].index
        for idx in index:
            x = portfolio.loc[idx]
            pos = x[x > 0].sum()
            neg = x[x < 0].sum()
            expo = pos + neg

            if expo > 0:
                to_pos = -neg + exposure
                multiplier = to_pos / pos
                x[x > 0] = x[x > 0] * multiplier
            elif expo < 0:
                to_neg = -pos - exposure
                multiplier = to_neg / neg
                x[x < 0] = x[x < 0] * multiplier

            rval.loc[idx] = x

        return rval

    def product_constraint_positions(self, portfolio, max_exposure=None):
        if max_exposure is None:
            max_exposure = self._max_product
        lower = (-max_exposure)
        upper = max_exposure

        return portfolio.clip(lower=lower, upper=upper)

    def turnover_winning_weighted_positions(self,
                                            portfolios,
                                            returns,
                                            winsize=None,
                                            commission_rate=None):
        if winsize is None:
            winsize = self._turnover_winnratio_winsize

        if commission_rate is None:
            commission_rate = self._commission_rate

        data = dict()
        for key, portfolio in portfolios.items():
            turnover = portfolio.diff(
                periods=1) - portfolio.shift(periods=1, freq=None) * returns
            turnover = turnover.abs()
            turnover = turnover.rolling(window=winsize).sum()
            return_net = calc_portfolio_net_return(
                portfolio, returns, commission_rate).sum(axis=1, skipna=True)
            return_net_sign = return_net
            return_net_sign[return_net > 0] = 1
            return_net_sign[return_net < 0] = -1
            winning_ratio = return_net_sign.rolling(window=winsize).mean()
            data[key] = turnover / (winning_ratio + 1e-16)
        data = pd.DataFrame(data)
        universe = pd.DataFrame(True, index=data.index, columns=data.columns)
        rank = cross_section_demean_rank(data, universe)

        rval = pd.DataFrame(0, index=returns.index, columns=returns.columns)
        for col in data.columns:
            rval += portfolios[col].mul(rank[col], axis=0)

        return rval

    def run(self, portfolio, returns=None):
        rval = portfolio.copy()
        if self._portfolio_exposure:
            rval = self.portfolio_constraint_positions(portfolio=portfolio)
        if self._product_exposure:
            rval = self.product_constraint_positions(portfolio=portfolio)
        if returns is not None:
            rval = self.turnover_winning_weighted_positions(portfolio, returns)
        return rval


class TurnoverRisk(object):

    def __init__(self, **kwargs):
        self._grade_step = kwargs['grade_step']

    def portfolio_positions(self, portfolio):
        rval = portfolio.copy()
        if self._grade_step > 0:
            rval = ts_grade(rval, self._grade_step)
        return rval

    def run(self, portfolio):
        return self.portfolio_positions(portfolio=portfolio)


class VolatilityRisk(object):

    def __init__(self, **kwargs):
        self._volatility_winsize = kwargs['volatility_winsize']
        self._volatility_base = kwargs['volatility_base']
        self._upper_limit = kwargs['upper_limit']
        self._lower_limit = kwargs['lower_limit']

    def return_vol_ratio(self, portfolio, returns, is_portfolio=True):
        data = portfolio.copy()
        product_return_gross = data.shift(periods=1) * returns
        portfolio_return_gross = product_return_gross.sum(
            axis=1, skipna=True) if is_portfolio else product_return_gross
        portfolio_return_std = ts_std_mean0(portfolio_return_gross,
                                            self._volatility_winsize)
        rval = 1 / portfolio_return_std / self._volatility_base
        rval.iloc[0:self._volatility_winsize] = 1
        rval[rval >= self._upper_limit] = self._upper_limit
        rval[rval <= self._lower_limit] = self._lower_limit

        return rval

    def run(self, portfolio, returns, method):
        rval = portfolio.copy()
        if method == 'portfolio':
            ratio = self.return_vol_ratio(portfolio=portfolio,
                                          returns=returns,
                                          is_portfolio=True)
            rval = rval.mul(ratio, axis=0)
        elif method == 'product':
            ratio = self.return_vol_ratio(portfolio=portfolio,
                                          returns=returns,
                                          is_portfolio=False)
            rval *= ratio
        return rval


#def inverse_average_ratio(close, high, low, average_winsize):
#    true_range = ts_true_range(c=close, h=high, l=low)
#    alpha = 1 / average_winsize
#    avg_true_range = BaseOperator.ts_wma(data=true_range, alpha=alpha)

#    numerator = 1 / (avg_true_range / low)
#    denominator = numerator.sum(axis=1, skipna=True)
#    rval = numerator.div(denominator, axis=0)
#    return rval



# -*- encoding:utf-8 -*-
"""
    LDA降维
"""
import numpy as np
import pandas as pd
from six.moves import xrange
from sklearn.pipeline import Pipeline
from sklearn.naive_bayes import GaussianNB
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.model_selection import cross_val_score
from sklearn.model_selection import RepeatedStratifiedKFold
from ultron.utilities.logger import kd_logger
from ultron.kdutils.progress import Progress


class LDA(object):

    def __init__(self, features=None):
        self._features = features

    def evaluate_model(self, model, X, y):
        cv = RepeatedStratifiedKFold(n_splits=10, n_repeats=3, random_state=1)
        scores = cross_val_score(model,
                                 X,
                                 y,
                                 scoring='accuracy',
                                 cv=cv,
                                 n_jobs=-1,
                                 error_score='raise')
        return scores

    def fetch_models(self, best_k):
        models = {}
        for i in range(1, best_k):
            steps = [('lda', LinearDiscriminantAnalysis(n_components=i)),
                     ('m', GaussianNB())]
            models[str(i)] = Pipeline(steps=steps)
        return models

    def fit_lda(self, X, y, best_k):
        results = []
        i = 0
        models = self.fetch_models(best_k=best_k)
        with Progress(len(models), 0, 'fit LDA') as pg:
            for name, model in models.items():
                i += 1
                scores = self.evaluate_model(model, X, y)
                results.append({
                    'name': name,
                    'mean': np.mean(scores),
                    'std': np.std(scores)
                })
                pg.show(i)
        results = pd.DataFrame(results).sort_values(by='mean', ascending=False)
        return int(results['name'].values[0])

    def run(self, factors_data, best_k):
        n_components = self.fit_lda(X=factors_data[self._features].values,
                                    y=factors_data['cluster'].values,
                                    best_k=best_k)
        dimension_factor = LinearDiscriminantAnalysis(
            n_components=n_components).fit_transform(
                factors_data[self._features].values,
                factors_data['cluster'].values,
            )
        dimension_name = [
            "lda_{0}_{1}".format(n_components, k)
            for k in xrange(1, n_components + 1)
        ]
        dimension_factor = pd.DataFrame(dimension_factor,
                                        columns=dimension_name)
        return dimension_factor



# -*- encoding:utf-8 -*-
"""示例ump主裁特征走势拟合角度模块"""

from ultron.ump.trade.ml_feature import FeatureDeg, FeatureDegExtend
from ultron.ump.ump.base import ump_main_make_xy, BuyUmpMixin
from ultron.ump.ump.main_base import MainBase
from ultron.ump.model.principles import Principles


class DegFiter(Principles):
    """
        内部类，只需要实现make_xy，且使用ump_main_make_xy装饰

        在__init__中通过：
                self.fiter_cls = self.get_fiter_class()
                self.fiter = self.fiter_cls(orders_pd=self.orders_pd, **kwarg)
        构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
    """

    @ump_main_make_xy
    def make_xy(self, **kwarg):
        """
        make_xy通过装饰器ump_main_make_xy进行二次包装
        这里只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df
        :param kwarg: ump_main_make_xy装饰器中使用kwarg
                      kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
                      kwargs['scaler']    做为可选关键字参数：控制在make_xy中返回的特征矩阵数据是否进行标准化处理

        :return: self.order_has_ret中使用filter选取特征列形成deg_df
                 ump_main_make_xy装饰器在make_xy返回deg_df后做转换matrix，形成x，y等工作

            eg: deg_df

                            result  buy_deg_ang42  buy_deg_ang252  buy_deg_ang60  \
            2014-09-24       0          3.378           3.458          3.458
            2014-10-24       0          0.191           2.889          2.809
            2014-10-29       1         -2.026          16.689         -0.761
            2014-10-29       1         -3.427         -11.956         -8.296
            2014-10-29       1         -2.915          39.469         -6.043
            2014-10-29       1         -2.026          16.689         -0.761
            2014-11-03       1          0.103          39.202         -4.614
            2014-11-11       1          8.341          -9.450          0.730
            2014-11-12       0          3.963           6.595         -7.524
            2014-11-26       0         14.052           6.061          7.566
            ...            ...            ...             ...            ...
            2016-03-14       1          4.002         -10.559         -7.992
            2016-03-14       0          0.129          -6.649        -10.880
            2016-03-30       0         13.121          -8.461          4.498
            2016-04-04       1          4.409         -33.097         -6.281
            2016-04-13       0          6.603         -31.459          0.191
            2016-04-14       0          4.611          18.428          3.134
            2016-04-15       0          4.238         -13.247          4.693
            2016-04-15       0          4.238         -13.247          4.693
            2016-04-29       1          1.445          16.266          4.615
            2016-04-29       1          1.445          16.266          4.615

                        buy_deg_ang21
            2014-09-24          1.818
            2014-10-24         -1.089
            2014-10-29          1.980
            2014-10-29          6.507
            2014-10-29          7.046
            2014-10-29          1.980
            2014-11-03         10.125
            2014-11-11         12.397
            2014-11-12          6.671
            2014-11-26         12.494
            ...                   ...
            2016-03-14          9.324
            2016-03-14          5.201
            2016-03-30          4.070
            2016-04-04          5.618
            2016-04-13          4.457
            2016-04-14          0.733
            2016-04-15          1.162
            2016-04-15          1.162
            2016-04-29         -1.115
            2016-04-29         -1.115
        """
        # regex='result|buy_deg_ang42|buy_deg_ang252|buy_deg_ang60|buy_deg_ang21'
        regex = 'result|{}'.format('|'.join(
            FeatureDeg().get_feature_ump_keys(ump_cls=MainDeg)))
        # noinspection PyUnresolvedReferences
        deg_df = self.order_has_ret.filter(regex=regex)
        return deg_df


class MainDeg(MainBase, BuyUmpMixin):
    """主裁走势拟合角度特征类，MainBase子类，混入BuyUmpMixin，做为买入ump类"""

    def get_predict_col(self):
        """
        主裁走势拟合角度特征keys：['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21']
        :return: ['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21']
        """
        return FeatureDeg().get_feature_ump_keys(ump_cls=MainDeg)

    def get_fiter_class(self):
        """
        主裁特征走势拟合角度返回的UltronMLPd子类：UltronUmpMainDeg.UmpDegFiter
        :return: UltronUmpMainDeg.UmpDegFiter
        """
        return DegFiter

    @classmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法：return 'deg_main'
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见UltronUmpManager中extend_ump_block方法
        """
        return 'deg_main'


class ExtendFeatureFiter(Principles):
    """角度主裁扩展类make_xy"""

    @ump_main_make_xy
    def make_xy(self, **kwarg):
        # 这里使用get_feature_ump_keys，只需要传递当前类名称即可，其根据是买入ump还是卖出ump返回对应特征列
        col = FeatureDegExtend().get_feature_ump_keys(ump_cls=MainDegExtend)
        regex = 'result|{}'.format('|'.join(col))
        extend_deg_df = self.order_has_ret.filter(regex=regex)
        return extend_deg_df


class MainDegExtend(MainBase, BuyUmpMixin):
    """主裁使用新的视角来决策交易，UltronUmpMainBase子类，混入BuyUmpMixin，做为买入ump类"""

    def get_predict_col(self):
        # 这里使用get_feature_ump_keys，只需要传递当前类名称即可，其根据是买入ump还是卖出ump返回对应特征列
        col = FeatureDegExtend().get_feature_ump_keys(ump_cls=MainDegExtend)
        return col

    def get_fiter_class(self):
        return ExtendFeatureFiter

    @classmethod
    def class_unique_id(cls):
        return 'extend_main_deg'
